# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Abstract mathematical operators."""

from __future__ import absolute_import, division, print_function

import inspect
import sys
from builtins import object
from numbers import Integral, Number

from odl.core.set import Field, LinearSpace, Set
from odl.core.set.space import LinearSpaceElement

__all__ = (
    'Operator',
    'OperatorComp',
    'OperatorSum',
    'OperatorVectorSum',
    'OperatorLeftScalarMult',
    'OperatorRightScalarMult',
    'FunctionalLeftVectorMult',
    'OperatorLeftVectorMult',
    'OperatorRightVectorMult',
    'OperatorPointwiseProduct',
    'OpTypeError',
    'OpDomainError',
    'OpRangeError',
    'OpNotImplementedError',
)


def _default_call_out_of_place(op, x, **kwargs):
    """Default out-of-place evaluation.

    Parameters
    ----------
    op : `Operator`
        Operator to call
    x : ``op.domain`` element
        Point in which to call the operator.
    kwargs:
        Optional arguments to the operator.

    Returns
    -------
    out : `range` element
        An object in the operator range. The result of an operator
        evaluation.
    """
    out = op.range.element()
    result = op._call_in_place(x, out, **kwargs)
    if result is not None and result is not out:
        raise ValueError('`op` returned a different value than `out`.'
                         'With in-place evaluation, the operator can '
                         'only return nothing (`None`) or the `out` '
                         'parameter.')
    return out


def _default_call_in_place(op, x, out, **kwargs):
    """Default in-place evaluation using ``Operator._call()``.

    Parameters
    ----------
    op : `Operator`
        Operator to call
    x : ``op.domain`` element
        Point in which to call the operator.
    out : ``op.range`` element
        An object in the operator range. The result of an operator
        evaluation is written here.
    kwargs:
        Optional arguments to the operator.
    """
    out.assign(op.range.element(op._call_out_of_place(x, **kwargs)))


def _function_signature(func):
    """Return the signature of a callable as a string.

    Parameters
    ----------
    func : callable
        Function whose signature to extract.

    Returns
    -------
    sig : string
        Signature of the function.
    """
    assert (sys.version_info.major > 2)
    return func.__name__ + str(inspect.signature(func))


def _dispatch_call_args(cls=None, bound_call=None, unbound_call=None,
                        attr='_call'):
    """Check the arguments of ``_call()`` or similar for conformity.

    The ``_call()`` method of `Operator` is allowed to have the
    following signatures:

        - ``_call(self, x)``
        - ``_call(self, vec, out)``
        - ``_call(self, x, out=None)``
        - ``_call(self, x, *, out=None)`` (``out`` as keyword-only
          argument)

    For disambiguation, the instance name (the first argument) **must**
    be 'self'.

    The name of the ``out`` argument **must** be 'out', the second
    argument may have any name.

    Additional variable ``**kwargs`` and keyword-only arguments
    are also allowed.

    Not allowed:
        - ``_call(self)`` -- No arguments except instance:
        - ``_call(x)`` -- 'self' missing, i.e. ``@staticmethod``
        - ``_call(cls, x)``  -- 'self' missing, i.e. ``@classmethod``
        - ``_call(self, out, x)`` -- ``out`` as second argument
        - ``_call(self, *x)`` -- Variable arguments
        - ``_call(self, x, y, out=None)`` -- more positional arguments
        - ``_call(self, x, out=False)`` -- default other than None for
          ``out``

    In particular, static or class methods are not allowed.

    Parameters
    ----------
    cls : `class`, optional
        The ``_call()`` method of this class is checked. If omitted,
        provide ``unbound_call`` instead to check directly.
    bound_call : callable, optional
        Check this bound method instead of ``cls``
    unbound_call : callable, optional
        Check this unbound function instead of ``cls``
    attr : string, optional
        Check this attribute instead of ``_call``, e.g. ``__call__``

    Returns
    -------
    has_out : bool
        Whether the call has an ``out`` argument
    out_is_optional : bool
        Whether the ``out`` argument is optional
    spec : `inspect.ArgSpec` or `inspect.FullArgSpec`
        Argument specification of the checked call function

    Raises
    ------
    ValueError
        if the signature of the function is malformed
    """
    py3 = (sys.version_info.major > 2)

    specs = ['_call(self, x[, **kwargs])',
             '_call(self, x, out[, **kwargs])',
             '_call(self, x, out=None[, **kwargs])']

    if py3:
        specs += ['_call(self, x, *, out=None[, **kwargs])']

    spec_msg = "\nPossible signatures are ('[, **kwargs]' means optional):\n\n"
    spec_msg += '\n'.join(specs)
    spec_msg += '\n\nStatic or class methods are not allowed.'

    if sum(arg is not None for arg in (cls, bound_call, unbound_call)) != 1:
        raise ValueError('exactly one object to check must be given')

    if cls is not None:
        # Get the actual implementation, including ancestors
        for parent in cls.mro():
            call = parent.__dict__.get(attr, None)
            if call is not None:
                break
        # Static and class methods are not allowed
        if isinstance(call, staticmethod):
            raise TypeError("'{}.{}' is a static method. "
                            "".format(cls.__name__, attr) + spec_msg)
        elif isinstance(call, classmethod):
            raise TypeError("'{}.{}' is a class method. "
                            "".format(cls.__name__, attr) + spec_msg)

    elif bound_call is not None:
        call = bound_call
        if not inspect.ismethod(call):
            raise TypeError('{} is not a bound method'.format(call))
    else:
        call = unbound_call

    if py3:
        # support kw-only args and annotations
        spec = inspect.getfullargspec(call)
        kw_only = spec.kwonlyargs
        kw_only_defaults = spec.kwonlydefaults
    else:
        spec = inspect.getargspec(call)
        kw_only = ()
        kw_only_defaults = {}

    signature = _function_signature(call)

    pos_args = spec.args
    if unbound_call is not None:
        # Add 'self' to positional arg list to satisfy the checker
        pos_args.insert(0, 'self')

    pos_defaults = spec.defaults
    varargs = spec.varargs

    # Variable args are not allowed
    if varargs is not None:
        raise ValueError("bad signature '{}': variable arguments not allowed"
                         "".format(signature) + spec_msg)

    if len(pos_args) not in (2, 3):
        raise ValueError("bad signature '{}'".format(signature) + spec_msg)

    true_pos_args = pos_args[1:]
    if len(true_pos_args) == 1:  # 'out' kw-only
        if 'out' in true_pos_args:  # 'out' positional and 'x' kw-only -> no
            raise ValueError("bad signature '{}': `out` cannot be the only "
                             "positional argument"
                             "".format(signature) + spec_msg)
        else:
            if 'out' not in kw_only:
                has_out = out_optional = False
            elif kw_only_defaults['out'] is not None:
                raise ValueError(
                    "bad signature '{}': `out` can only default to "
                    "`None`, got '{}'"
                    "".format(signature, kw_only_defaults['out']) +
                    spec_msg)
            else:
                has_out = True
                out_optional = True

    elif len(true_pos_args) == 2:  # Both args positional
        if true_pos_args[0] == 'out':  # out must come second
            py3_txt = ' or keyword-only. ' if py3 else '. '
            raise ValueError("bad signature '{}': `out` can only be the "
                             "second positional argument".format(signature) +
                             py3_txt + spec_msg)
        elif true_pos_args[1] != 'out':  # 'out' must be 'out'
            raise ValueError("bad signature '{}': output parameter must "
                             "be called 'out', got '{}'"
                             "".format(signature, true_pos_args[1]) +
                             spec_msg)
        else:
            has_out = True
            out_optional = bool(pos_defaults)
            if pos_defaults and pos_defaults[-1] is not None:
                raise ValueError("bad signature '{}': `out` can only "
                                 "default to `None`, got '{}'"
                                 "".format(signature, pos_defaults[-1]) +
                                 spec_msg)

    else:  # Too many positional args
        raise ValueError("bad signature '{}': too many positional arguments"
                         " ".format(signature) + spec_msg)

    return has_out, out_optional, spec


class Operator(object):

    r"""Abstract mathematical operator.

    An operator is a mapping

    .. math::
        \mathcal{A}: \mathcal{X} \to \mathcal{Y}

    between sets :math:`\mathcal{X}` (domain) and :math:`\mathcal{Y}`
    (range). The evaluation of :math:`\mathcal{A}` at an element
    :math:`x \in \mathcal{X}` is denoted by :math:`\mathcal{A}(x)`
    and produces an element in :math:`\mathcal{Y}`:

    .. math::
        y = \mathcal{A}(x) \in \mathcal{Y}.

    Programmatically, these properties are reflected in the `Operator`
    class described in the following.

    **Abstract attributes and methods**

    `Operator` is an **abstract** class, i.e. it can only be
    subclassed, not used directly.

    Any subclass of `Operator` must have the following
    attributes:

    ``domain`` : `Set`
        The set of elements this operator can be applied to

    ``range`` : `Set`
        The set this operator maps to

    It is **highly** recommended to call
    ``super(MyOp, self).__init__(domain, range)``  in the ``__init__()``
    method of any subclass  ``MyOp``, where ``domain`` and ``range`` are
    the arguments specifying domain and range of the new operator. In that
    case, the attributes `Operator.domain` and `Operator.range` are
    automatically provided by the parent class `Operator`.

    In addition, any subclass **must** implement the private method
    `Operator._call()`. It signature determines how it is interpreted:


    **In-place-only evaluation:** ``_call(self, x, out[, **kwargs])``

    In-place evaluation means that the operator is applied, and the
    result is written to an existing element ``out`` provided,
    i.e.

        ``_call(self, x, out)  <==>  out <-- operator(x)``

    **Parameters:**

    x : `Operator.domain` element
        An object in the operator domain to which the operator is
        applied

    out : `Operator.range` element
        An object in the operator range to which the result of the
        operator evaluation is written.

    **Returns:**

    None (return value is ignored)


    **Out-of-place-only evaluation:** ``_call(self, x[, **kwargs])``

    Out-of-place evaluation means that the operator is applied,
    and the result is written to a **new** element which is returned.
    In this case, a subclass has to implement the method

        ``_call(self, x)  <==>  operator(x)``

    **Parameters:**

    x : `Operator.domain` element
        An object in the operator domain to which the operator is
        applied

    **Returns:**

    out : `Operator.range` `element-like`
        An object in the operator range holding the result of the
        operator evaluation


    **Dual-use evaluation:** ``_call(self, x, out=None[, **kwargs])``

    Evaluate in-place if ``out`` is given, otherwise out-of-place.

    **Parameters:**

    x : `Operator.domain` element
        An object in the operator domain to which the operator is
        applied

    out : `Operator.range` element, optional
        An object in the operator range to which the result of the
        operator evaluation is written

    **Returns:**

    None (return value is ignored)


    Notes
    -----
    - If `Operator._call` is implemented in-place-only or
      out-of-place-only and the `Operator.range` is a `LinearSpace`,
      a default implementation of the respective other is provided.

    - `Operator._call` is allowed to have keyword-only arguments (Python
      3 only).

    - The term "element-like" means that an object must be convertible
      to an element by the ``domain.element()`` method.
    """

    def __new__(cls, *args, **kwargs):
        """Create a new instance."""
        if '_call_out_of_place' not in cls.__dict__:
            call_has_out, call_out_optional, _ = _dispatch_call_args(cls)
            cls._call_has_out = call_has_out
            cls._call_out_optional = call_out_optional
            if not call_has_out:
                # Out-of-place _call
                cls._call_in_place = _default_call_in_place
                cls._call_out_of_place = cls._call
            elif call_out_optional:
                # Dual-use _call
                cls._call_in_place = cls._call_out_of_place = cls._call
            else:
                # In-place-only _call
                cls._call_in_place = cls._call
                cls._call_out_of_place = _default_call_out_of_place

        return object.__new__(cls)

    def __init__(self, domain, range, linear=False):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `Set`
            The domain of this operator, i.e., the set of elements to
            which this operator can be applied
        range : `Set`
            The range of this operator, i.e., the set this operator
            maps to
        linear : bool, optional
            If ``True``, the operator is considered as linear. In this
            case, ``domain`` and ``range`` have to be instances of
            `LinearSpace`, or `Field`.
        """
        if not isinstance(domain, Set):
            raise TypeError('`domain` {!r} is not a `Set` instance'
                            ''.format(domain))
        if not isinstance(range, Set):
            raise TypeError('`range` {!r} is not a `Set` instance'
                            ''.format(range))

        self.__domain = domain
        self.__range = range
        self.__is_linear = bool(linear)

        # Cache for efficiency since this is done in each call.
        self.__is_functional = isinstance(range, Field)

        # Mandatory out makes no sense for functionals.
        # However, we need to allow optional out to support vectorized
        # functions (which are functionals in the duck-typing sense).
        if (self.is_functional and self._call_has_out and
                not self._call_out_optional):
            raise ValueError('mandatory `out` parameter not allowed for '
                             'functionals')

        if self.is_linear:
            if not isinstance(domain, (LinearSpace, Field)):
                raise TypeError('`domain` {!r} not a `LinearSpace` or `Field` '
                                'instance'.format(domain))
            if not isinstance(range, (LinearSpace, Field)):
                raise TypeError('`range` {!r} not a `LinearSpace` or `Field` '
                                'instance'.format(range))

    def _call(self, x, out=None, **kwargs):
        """Implementation of the operator evaluation.

        This method is the private backend for the evaluation of an
        operator. It needs to match certain signature conventions,
        and its implementation type is inferred from its signature.

        The following signatures are allowed:

            - ``_call(self, x)``  -->  out-of-place evaluation
            - ``_call(self, vec, out)``  -->  in-place evaluation
            - ``_call(self, x, out=None)``   --> both
            - ``_call(self, x, *, out=None)`` (``out`` as keyword-only
              argument)  --> both

        For disambiguation, the instance name (the first argument) **must**
        be 'self'.

        The name of the ``out`` argument **must** be 'out', the second
        argument may have any name.

        Additional variable ``**kwargs`` and keyword-only arguments
        are also allowed.

        Notes
        -----
        Some general advice on how to implement operator evaluation:

        - If you just write a quick implementation or are not too
          worried about efficiency, it may be easiest to write the
          evaluation *out-of-place*.
        - We recommend advanced and performance-aware users to implement
          the *in-place* pattern if the wrapped code supports it.
          In-place evaluation is usually significantly faster since it
          avoids the allocation of new memory and a copy compared to
          out-of-place evaluation.
        - If there is a significant performance gain from implementing
          an out-of-place method separately, use the pattern for both
          (``out`` optional) and decide according to the given ``out``
          parameter which one to use.
        - If your evaluation code does not support in-place evaluation,
          use the out-of-place pattern.

        Note that the public call pattern ``op()`` using ``op.__call__``
        provides a default implementation of the underlying in-place or
        out-of-place call even if you choose the respective other
        pattern.

        See the `operator guide
        <https://odlgroup.github.io/odl/guide/in_depth/operator_guide.html>`_
        for more info on in-place vs. out-of-place evaluation.

        Parameters
        ----------
        x : `domain` `element-like`
            Element to which the operator is applied
        out : `range` element, optional
            Element to which the result is written

        Returns
        -------
        out : `range` `element-like`
            Result of the evaluation. If ``out`` was provided, the
            returned object is a reference to it.
        """
        raise NotImplementedError('this operator {!r} does not implement '
                                  '`_call`. See `Operator._call` for '
                                  'instructions on how to do this.'
                                  ''.format(self))

    @property
    def domain(self):
        """Set of objects on which this operator can be evaluated."""
        return self.__domain

    @property
    def range(self):
        """Set in which the result of an evaluation of this operator lies."""
        return self.__range

    @property
    def is_linear(self):
        """``True`` if this operator is linear."""
        return self.__is_linear

    @property
    def is_functional(self):
        """``True`` if this operator's range is a `Field`."""
        return self.__is_functional

    @property
    def adjoint(self):
        """Adjoint of this operator (abstract).

        Raises
        ------
        OpNotImplementedError
            Since the adjoint cannot be default implemented.
        """
        raise OpNotImplementedError('adjoint not implemented '
                                    'for operator {!r}'
                                    ''.format(self))

    def derivative(self, point):
        """Return the operator derivative at ``point``.

        Raises
        ------
        OpNotImplementedError
            If the operator is not linear, the derivative cannot be
            default implemented.
        """
        if self.is_linear:
            return self
        else:
            raise OpNotImplementedError('derivative not implemented '
                                        'for operator {!r}'
                                        ''.format(self))

    @property
    def inverse(self):
        """Return the operator inverse.

        Raises
        ------
        OpNotImplementedError
            Since the inverse cannot be default implemented.
        """
        raise OpNotImplementedError('inverse not implemented for operator {!r}'
                                    ''.format(self))

    def __call__(self, x, out=None, **kwargs):
        """Return ``self(x[, out, **kwargs])``.

        Implementation of the call pattern ``op(x)`` with the private
        ``_call()`` method and added error checking.

        Parameters
        ----------
        x : `domain` `element-like`
            An object which can be converted into an element of this
            operator's domain with the ``self.domain.element`` method.
            The operator is applied to this object, which is treated
            as immutable, hence it is not modified during evaluation.
        out : `range` element, optional
            An object in the operator range to which the result of the
            operator evaluation is written. The result is independent
            of the initial state of this object.
        kwargs :
            Passed on to the underlying implementation in `_call`.

        Returns
        -------
        out : `range` element
            Result of the operator evaluation. If ``out`` was provided,
            the returned object is a reference to it.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> op = odl.ScalingOperator(rn, 2.0)
        >>> x = rn.element([1, 2, 3])

        Out-of-place evaluation:

        >>> op(x)
        rn(3).element([ 2.,  4.,  6.])

        In-place evaluation:

        >>> y = rn.element()
        >>> op(x, out=y)
        rn(3).element([ 2.,  4.,  6.])
        >>> y
        rn(3).element([ 2.,  4.,  6.])

        See Also
        --------
        _call : Implementation of the method
        """
        if x not in self.domain:
            try:
                x = self.domain.element(x)
            except (TypeError, ValueError):
                raise OpDomainError(
                    'unable to cast {!r} to an element of '
                    'the domain {!r}'.format(x, self.domain))

        if out is not None:  # In-place evaluation
            if out not in self.range:
                raise OpRangeError('`out` {!r} not an element of the range '
                                   '{!r} of {!r}'
                                   ''.format(out, self.range, self))

            if self.is_functional:
                raise TypeError('`out` parameter cannot be used '
                                'when range is a field')

            result = self._call_in_place(x, out=out, **kwargs)
            # TODO: At present, we perform an equality check on the entire array, which is as inefficient as it gets. We'd rather perform a pointer equality with the "is" keyword. However, the current machinery for the _call_in_place function might be creating a new out object, which leads to the "is" equality failing. We must investigate this _call_in_place function to identify when and why are objects created/deleted.
            if result is not None and result != out:
                raise ValueError('`op` returned a different value than `out`. '
                                 'With in-place evaluation, the operator can '
                                 'only return nothing (`None`) or the `out` '
                                 'parameter.')

        else:  # Out-of-place evaluation
            out = self._call_out_of_place(x, **kwargs)

            if self.domain is not None and out is None:
                raise OpRangeError(
                        "The out-of-place version of the operator does not return a value.")
            if out not in self.range:
                try:
                    out = self.range.element(out)
                except (TypeError, ValueError):
                    raise OpRangeError(
                        'unable to cast {!r} to an element of '
                        'the range {!r}'.format(out, self.range))
        return out

    def norm(self, estimate=False, **kwargs):
        """Return the operator norm of this operator.

        If this operator is non-linear, this should be the Lipschitz constant.

        Parameters
        ----------
        estimate : bool
            If true, estimate the operator norm. By default, it is estimated
            using `power_method_opnorm`, which is only applicable for linear
            operators.
            Subclasses are allowed to ignore this parameter if they can provide
            an exact value.

        Other Parameters
        ----------------
        kwargs :
            If ``estimate`` is True, pass these arguments to the
            `power_method_opnorm` call.

        Returns
        -------
        norm : float

        Examples
        --------
        Some operators know their own operator norm and do not need an estimate

        >>> spc = odl.rn(3)
        >>> id = odl.IdentityOperator(spc)
        >>> id.norm(True)
        1.0

        For others, there is no closed form expression and an estimate is
        needed:

        >>> spc = odl.uniform_discr(0, 1, 3)
        >>> grad = odl.Gradient(spc)
        >>> opnorm = grad.norm(estimate=True)
        """
        if not estimate:
            raise NotImplementedError('`Operator.norm()` not implemented, use '
                                      '`Operator.norm(estimate=True)` to '
                                      'obtain an estimate.')
        else:
            norm = getattr(self, '__norm', None)
            if norm is not None:
                return norm
            else:
                from odl.core.operator.oputils import power_method_opnorm
                self.__norm = power_method_opnorm(self, **kwargs)
                return self.__norm

    def __add__(self, other):
        """Return ``self + other``.

        If other is an element in self.range, this corresponds to

            ``self + other <==> (x --> self(x) + other)``

        If other is an element in self.range.field, this corresponds to

            ``self + other <==> (x --> self(x) + 1*other)``

        where ``1`` is the one-element in self.range.

        If other is an opertor, this corresponds to

            ``self + other <==> (x --> self(x) + other(x))``
        """
        if other in self.range:
            return OperatorVectorSum(self, other)
        elif other in self.range.field:
            constant_vector = other * self.range.one()
            return OperatorVectorSum(self, constant_vector)
        elif isinstance(other, Operator):
            return OperatorSum(self, other)
        else:
            return NotImplemented

    def __radd__(self, other):
        """Return ``other + self``."""
        # Use commutativity
        return self + other

    def __sub__(self, other):
        """Return ``self - other``."""
        return self + (-1) * other

    def __rsub__(self, other):
        """Return ``other - self``."""
        return (-1) * self + other

    def __mul__(self, other):
        """Return ``self * other``.

        If ``other`` is an operator, this corresponds to
        operator composition:

            ``(left * right)(x) == left(right(x))``

        If ``other`` is a scalar, this corresponds to scalar multiplication
        with the operator argument:

            ``(op * scalar)(x) == op(scalar * x)``

        If ``other`` is an ``op.domain`` element, this corresponds to
        vector multiplication with the operator argument:

            ``(op * y)(x) == op(y * x)``

        Note that left and right multiplications are generally
        different.

        Parameters
        ----------
        other : `Operator`, `LinearSpaceElement` or scalar
            `Operator`:
                The `Operator.domain` of ``other`` must match this
                operator's `Operator.range`.

            `LinearSpaceElement`:
                ``other`` must be an element of this operator's
                `Operator.domain`.

            scalar:
                The `Operator.domain` of this operator must be a
                `LinearSpace` and ``other`` must be an
                element of the ``field`` of this operator's
                `Operator.domain`.

        Returns
        -------
        mul : `Operator`
            The multiplication operator.

            If ``other`` is an operator, ``mul`` is an
            `OperatorComp`.

            If ``other`` is a scalar, ``mul`` is an
            `OperatorRightScalarMult`.

            If ``other`` is an ``op.domain`` element, ``mul`` is an
            `OperatorRightVectorMult`.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> op = odl.IdentityOperator(rn)
        >>> x = rn.element([1, 2, 3])
        >>> op(x)
        rn(3).element([ 1.,  2.,  3.])
        >>> Scaled = op * 3
        >>> Scaled(x)
        rn(3).element([ 3.,  6.,  9.])
        """
        if isinstance(other, Operator):
            return OperatorComp(self, other)
        elif isinstance(other, Number):
            # Left multiplication is more efficient, so we can use this in the
            # case of linear operator.
            if self.is_linear:
                return other * self
            else:
                return OperatorRightScalarMult(self, other)
        elif isinstance(other, LinearSpaceElement) and other in self.domain:
            return OperatorRightVectorMult(self, other.copy())
        else:
            return NotImplemented

    def __matmul__(self, other):
        """Return ``self @ other``.

        See `Operator.__mul__`
        """
        return self.__mul__(other)

    def __rmul__(self, other):
        """Return ``other * self``.

        If ``other`` is an `Operator`, this corresponds to
        operator composition:

            ``(left * right)(x) == left(right(x))``

        If ``other`` is a scalar, this corresponds to scalar multiplication
        with the operator evaluation result:

            ``(scalar * op)(x) == scalar * op(x)``

        If ``other`` is an ``op.domain`` element, this corresponds to
        vector multiplication with the operator evaluation result:

            ``(y * op)(x) == y * op(x)``

        Note that left and right multiplications are generally
        different.

        Parameters
        ----------
        other : {`Operator`, `LinearSpaceElement`, scalar}
            `Operator`:
                The `Operator.range` of ``other`` must match this
                operator's `Operator.domain`

            `LinearSpaceElement`:
                ``other`` must be an element of `Operator.range`.

            scalar:
                `Operator.range` must be a
                `LinearSpace` and ``other`` must be an
                element of ``self.range.field``.

        Returns
        -------
        mul : `Operator`
            The multiplication operator.

            If ``other`` is an operator, ``mul`` is an
            `OperatorComp`.

            If ``other`` is a scalar, ``mul`` is an
            `OperatorLeftScalarMult`.

            If ``other`` is an ``op.range`` element, ``mul`` is an
            `OperatorLeftVectorMult`.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> op = odl.IdentityOperator(rn)
        >>> x = rn.element([1, 2, 3])
        >>> op(x)
        rn(3).element([ 1.,  2.,  3.])
        >>> Scaled = 3 * op
        >>> Scaled(x)
        rn(3).element([ 3.,  6.,  9.])
        """
        if isinstance(other, Operator):
            return OperatorComp(other, self)
        elif isinstance(other, Number):
            return OperatorLeftScalarMult(self, other)
        elif other in self.range:
            return OperatorLeftVectorMult(self, other.copy())
        elif (isinstance(other, LinearSpaceElement) and
              other.space.field == self.range):
            return FunctionalLeftVectorMult(self, other.copy())
        else:
            return NotImplemented

    def __rmatmul__(self, other):
        """Return ``other @ op``.

        See `Operator.__rmul__`
        """
        return self.__rmul__(other)

    def __pow__(self, n):
        """Return ``op ** n``.

        This corresponds to the power of an operator:

            ``(op ** 1)(x) == op(x)``
            ``(op ** 2)(x) == op(op(x))``
            ``(op ** 3)(x) == op(op(op(x)))``
            ...

        Parameters
        ----------
        n : positive int
            The power the operator should be taken to.

        Returns
        -------
        pow : `Operator`
            The power of this operator. If ``n == 1``, ``pow`` is
            this operator, for ``n > 1``, an `OperatorComp`

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> op = odl.ScalingOperator(rn, 3)
        >>> x = rn.element([1, 2, 3])
        >>> op(x)
        rn(3).element([ 3.,  6.,  9.])
        >>> squared = op ** 2
        >>> squared(x)
        rn(3).element([  9.,  18.,  27.])
        >>> squared = op**3
        >>> squared(x)
        rn(3).element([ 27.,  54.,  81.])
        """
        if isinstance(n, Integral) and n > 0:
            op = self
            while n > 1:
                op = OperatorComp(self, op)
                n -= 1
            return op
        else:
            return NotImplemented

    def __truediv__(self, other):
        """Return ``self / other``.

        If ``other`` is a scalar, this corresponds to right
        division of operators with scalars:

            ``(op / scalar)(x) == op(x / scalar)``

        Parameters
        ----------
        other : scalar
            If `Operator.range` is a `LinearSpace`, ``scalar`` must be
            an element of the ``field`` of this operator's range.

        Returns
        -------
        rmul : `OperatorRightScalarMult`
            The 'divided' operator.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> op = odl.IdentityOperator(rn)
        >>> x = rn.element([3, 6, 9])
        >>> op(x)
        rn(3).element([ 3.,  6.,  9.])
        >>> Scaled = op / 3.0
        >>> Scaled(x)
        rn(3).element([ 1.,  2.,  3.])
        """
        if isinstance(other, Number):
            return self * (1.0 / other)
        else:
            return NotImplemented

    __div__ = __truediv__

    def __neg__(self):
        """Return ``-self``."""
        return -1 * self

    def __pos__(self):
        """Return ``+op``.

        The operator itself.
        """
        return self

    def __repr__(self):
        """Return ``repr(self)``.

        The default `repr` implementation. Should be overridden by
        subclasses.
        """
        return '{}: {!r} -> {!r}'.format(self.__class__.__name__, self.domain,
                                         self.range)

    def __str__(self):
        """Return ``str(self)``.

        The default string implementation. Should be overridden by
        subclasses.
        """
        return self.__class__.__name__

    # Give a `Operator` a higher priority than any NumPy array type. This
    # forces the usage of `__op__` of `Operator` if the other operand
    # is a NumPy object (applies also to scalars!).
    # Set higher than LinearSpaceElement.__array_priority__ to handle
    # vector multiplication properly
    __array_priority__ = 2000000.0


class OperatorSum(Operator):

    """Expression type for the sum of operators.

        ``OperatorSum(left, right)(x) == left(x) + right(x)``

    The sum is only well-defined for `Operator` instances where
    `Operator.range` is a `LinearSpace`.

    """

    def __init__(self, left, right, tmp_ran=None, tmp_dom=None):
        """Initialize a new instance.

        Parameters
        ----------
        left : `Operator`
            First summand. Its `Operator.range` must be a `LinearSpace` or
            `Field`.
        right : `Operator`
            Second summand. Must have the same `Operator.domain` and
            `Operator.range` as ``left``.
        tmp_ran : `Operator.range` element, optional
            Used to avoid the creation of a temporary when applying the
            operator.
        tmp_dom : `Operator.domain` element, optional
            Used to avoid the creation of a temporary when applying the
            operator adjoint.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> op = odl.IdentityOperator(r3)
        >>> x = r3.element([1, 2, 3])
        >>> out = r3.element()
        >>> OperatorSum(op, op)(x, out)  # In-place, returns out
        rn(3).element([ 2.,  4.,  6.])
        >>> out
        rn(3).element([ 2.,  4.,  6.])
        >>> OperatorSum(op, op)(x)
        rn(3).element([ 2.,  4.,  6.])
        """
        if left.range != right.range:
            raise OpTypeError('operator ranges {!r} and {!r} do not match'
                              ''.format(left.range, right.range))
        if not isinstance(left.range, (LinearSpace, Field)):
            raise OpTypeError('`left.range` {!r} not a `LinearSpace` or '
                              '`Field` instance'.format(left.range))
        if left.domain != right.domain:
            raise OpTypeError('operator domains {!r} and {!r} do not match'
                              ''.format(left.domain, right.domain))

        if tmp_ran is not None and tmp_ran not in left.range:
            raise OpRangeError('`tmp_ran` {!r} not an element of the operator '
                               'range {!r}'.format(tmp_ran, left.range))
        if tmp_dom is not None and tmp_dom not in left.domain:
            raise OpDomainError('`tmp_dom` {!r} not an element of the '
                                'operator domain {!r}'
                                ''.format(tmp_dom, left.domain))

        super(OperatorSum, self).__init__(
            left.domain, left.range,
            linear=left.is_linear and right.is_linear)
        self.__left = left
        self.__right = right
        self.__tmp_ran = tmp_ran
        self.__tmp_dom = tmp_dom

    @property
    def left(self):
        """The left/first part of this sum."""
        return self.__left

    @property
    def right(self):
        """The left/second part of this sum."""
        return self.__right

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.left(x) + self.right(x)
        else:
            tmp = (self.__tmp_ran if self.__tmp_ran is not None
                   else self.range.element())
            # Write to `tmp` first, otherwise aliased `x` and `out` lead
            # to wrong result
            self.left(x, out=tmp)
            self.right(x, out=out)
            out += tmp

    def derivative(self, x):
        """Return the operator derivative at ``x``.

        The derivative of a sum of two operators is equal to the sum of
        the derivatives.

        Parameters
        ----------
        x : `domain` `element-like`
            Evaluation point of the derivative
        """
        if self.is_linear:
            return self
        else:
            return OperatorSum(self.left.derivative(x),
                               self.right.derivative(x),
                               self.__tmp_dom, self.__tmp_ran)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator sum is the sum of the operator
        adjoints:

            ``OperatorSum(left, right).adjoint ==
            OperatorSum(left.adjoint, right.adjoint)``

        Returns
        -------
        adjoint : `OperatorSum`

        Raises
        ------
        OpNotImplementedError
            If either of the underlying operators are non-linear.
        """
        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        return OperatorSum(self.left.adjoint, self.right.adjoint,
                           self.__tmp_dom, self.__tmp_ran)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.left, self.right)

    def __str__(self):
        """Return ``str(self)``."""
        return '({} + {})'.format(self.left, self.right)


class OperatorVectorSum(Operator):

    """Operator that computes ``op(x) + y``.

        ``OperatorVectorSum(op, y)(x) == op(x) + y``
    """

    def __init__(self, operator, vector):
        """Initialize a new instance.

        Parameters
        ----------
        operator : `Operator`
            Operator to be used in the sum. Its
            `Operator.range` must be a `LinearSpace`.
        vector : ``operator.range`` `element-like`
            Vector to be added to the operator result.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> y = r3.element([1, 2, 3])
        >>> ident_op = odl.IdentityOperator(r3)
        >>> sum_op = odl.OperatorVectorSum(ident_op, y)
        >>> x = r3.element([4, 5, 6])
        >>> sum_op(x)
        rn(3).element([ 5.,  7.,  9.])
        """
        if not isinstance(operator, Operator):
            raise TypeError('`op` {!r} not a Operator instance'
                            ''.format(operator))

        if not isinstance(operator.range, LinearSpace):
            raise TypeError('`op.range` {!r} not a LinearSpace instance'
                            ''.format(operator.range))

        super(OperatorVectorSum, self).__init__(
            operator.domain, operator.range)
        self.__operator = operator
        self.__vector = operator.range.element(vector)

    @property
    def operator(self):
        """The operator to apply."""
        return self.__operator

    @property
    def vector(self):
        """The constant operator range element to add."""
        return self.__vector

    def _call(self, x, out=None):
        """Evaluate the residual at ``x`` and write to ``out`` if given."""
        if out is None:
            out = self.operator(x)
        else:
            self.operator(x, out=out) 

        out += self.vector
        return out

    def derivative(self, point):
        """Derivative the operator vector sum.

        It is equal to the derivative of the "inner" operator:

            ``OperatorVectorSum(op, y).derivative(z) == op.derivative(z)``

        Parameters
        ----------
        point : `domain` element
            Any element in the domain where the derivative should be taken

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> op = odl.IdentityOperator(r3)
        >>> sum = odl.OperatorVectorSum(op, r3.element([1, 2, 3]))
        >>> x = r3.element([4, 5, 6])
        >>> sum.derivative(x)(x)
        rn(3).element([ 4.,  5.,  6.])
        """
        return self.operator.derivative(point)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        # return '({} + {})'.format(self.left, self.right)
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.vector)


class OperatorComp(Operator):

    """Expression type for the composition of operators.

        ``OperatorComp(left, right)(x) == left(right(x))``

    The composition is only well-defined if ``left.domain == right.range``.
    """

    def __init__(self, left, right, tmp=None):
        """Initialize a new `OperatorComp` instance.

        Parameters
        ----------
        left : `Operator`
            The left ("outer") operator
        right : `Operator`
            The right ("inner") operator. Its range must coincide with the
            domain of ``left``.
        tmp : element of the range of ``right``, optional
            Used to avoid the creation of a temporary when applying the
            operator.
        """
        if right.range != left.domain:
            raise OpTypeError('`range` {!r} of the right operator {!r} not '
                              'equal to the domain {!r} of the left '
                              'operator {!r}'
                              ''.format(right.range, right,
                                        left.domain, left))

        if tmp is not None and tmp not in left.domain:
            raise OpDomainError('`tmp` {!r} not an element of the left '
                                'operator domain {!r}'
                                ''.format(tmp, left.domain))

        super(OperatorComp, self).__init__(
            right.domain, left.range,
            linear=left.is_linear and right.is_linear)
        self.__left = left
        self.__right = right
        self.__tmp = tmp

    @property
    def left(self):
        """The left/first part of this composition."""
        return self.__left

    @property
    def right(self):
        """The left/second part of this composition."""
        return self.__right

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.left(self.right(x))
        else:
            tmp = (self.__tmp if self.__tmp is not None
                   else self.right.range.element())
            self.right(x, out=tmp)
            return self.left(tmp, out=out)

    @property
    def inverse(self):
        """Inverse of this operator.

        The inverse of the operator composition is the composition of
        the inverses in reverse order:

            ``OperatorComp(left, right).inverse ==``
            ``OperatorComp(right.inverse, left.inverse)``
        """
        return OperatorComp(self.right.inverse, self.left.inverse,
                            self.__tmp)

    def derivative(self, x):
        """Return the operator derivative.

        The derivative of the operator composition follows the chain
        rule:

            ``OperatorComp(left, right).derivative(y) ==
            OperatorComp(left.derivative(right(y)), right.derivative(y))``

        Parameters
        ----------
        x : `domain` `element-like`
            Evaluation point of the derivative. Needs to be usable as
            input for the ``right`` operator.
        """
        if self.is_linear:
            return self
        else:
            if self.left.is_linear:
                left_deriv = self.left
            else:
                left_deriv = self.left.derivative(self.right(x))

            right_deriv = self.right.derivative(x)

            return OperatorComp(left_deriv, right_deriv,
                                self.__tmp)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator composition is the composition of
        the operator adjoints in reverse order:

            ``OperatorComp(left, right).adjoint ==
            OperatorComp(right.adjoint, left.adjoint)``

        Returns
        -------
        adjoint : `OperatorComp`

        Raises
        ------
        OpNotImplementedError
            If any of the underlying operators are non-linear.
        """
        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        return OperatorComp(self.right.adjoint, self.left.adjoint,
                            self.__tmp)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.left, self.right)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} o {}'.format(self.left, self.right)


class OperatorPointwiseProduct(Operator):

    """Expression type for the pointwise operator mulitplication.

        ``OperatorPointwiseProduct(left, right)(x) == left(x) * right(x)``
    """

    def __init__(self, left, right):
        """Initialize a new instance.

        Parameters
        ----------
        left : `Operator`
            The first factor
        right : `Operator`
            The second factor. Must have the same domain and range as
            ``left``.
        """
        if left.range != right.range:
            raise OpTypeError('operator ranges {!r} and {!r} do not match'
                              ''.format(left.range, right.range))
        if not isinstance(left.range, (LinearSpace, Field)):
            raise OpTypeError('range {!r} not a `LinearSpace` or `Field` '
                              'instance'.format(left.range))
        if left.domain != right.domain:
            raise OpTypeError('operator domains {!r} and {!r} do not match'
                              ''.format(left.domain, right.domain))

        super(OperatorPointwiseProduct, self).__init__(
            left.domain, left.range, linear=False)
        self.__left = left
        self.__right = right

    @property
    def left(self):
        """The left/first part of this multiplication."""
        return self.__left

    @property
    def right(self):
        """The left/second part of this multiplication."""
        return self.__right

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.left(x) * self.right(x)
        else:
            tmp = self.right.range.element()
            # Write to `tmp` first, otherwise aliased `x` and `out` lead
            # to wrong result
            self.left(x, out=tmp)
            self.right(x, out=out)
            out *= tmp

    def derivative(self, x):
        """Return the derivative at ``x``."""
        if self.is_linear:
            return self
        else:
            left = self.right(x) @ self.left.derivative(x)
            right = self.left(x) @ self.right.derivative(x)
            return left + right

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.left, self.right)

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


class OperatorLeftScalarMult(Operator):

    """Expression type for the operator left scalar multiplication.

        ``OperatorLeftScalarMult(op, s)(x) == s * op(x)``

    The scalar multiplication is well-defined only if ``op.range`` is
    a `LinearSpace`.
    """

    def __init__(self, operator, scalar):
        """Initialize a new `OperatorLeftScalarMult` instance.

        Parameters
        ----------
        operator : `Operator`
            Operator in the scalar multiplication. Its `range` must be
            a `LinearSpace` or `Field`.
        scalar : ``operator.range.field`` element
            A real or complex number, depending on the field of
            the range.

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorLeftScalarMult(operator, 3)
        >>> left_mul_op([1, 2, 3])
        rn(3).element([ 3.,  6.,  9.])
        """
        if not isinstance(operator.range, (LinearSpace, Field)):
            raise OpTypeError('range {!r} not a `LinearSpace` or `Field` '
                              'instance'.format(operator.range))

        if scalar not in operator.range.field:
            raise TypeError('`scalar` {!r} not in the field {!r} of the '
                            'operator range {!r}'
                            ''.format(scalar,
                                      operator.range.field,
                                      operator.range))

        if isinstance(operator, OperatorLeftScalarMult):
            # Shortcut to save performance in case of repeated multiplications
            scalar = scalar * operator.scalar
            operator = operator.operator

        super(OperatorLeftScalarMult, self).__init__(
            operator.domain, operator.range, linear=operator.is_linear)
        self.__operator = operator
        self.__scalar = scalar

    @property
    def operator(self):
        """The operator part of this multiplication."""
        return self.__operator

    @property
    def scalar(self):
        """The scalar part of this multiplication."""
        return self.__scalar

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.scalar * self.operator(x)
        else:
            self.operator(x, out=out)
            out *= self.scalar

    @property
    def inverse(self):
        """Inverse of this operator.

        The inverse of ``scalar * op`` is given by
        ``op.inverse * 1/scalar`` if ``scalar != 0``. If ``scalar == 0``,
        the inverse is not defined.

            ``OperatorLeftScalarMult(op, s).inverse ==
            OperatorRightScalarMult(op.inverse, 1/s)``

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorLeftScalarMult(operator, 3)
        >>> left_mul_op.inverse([3, 3, 3])
        rn(3).element([ 1.,  1.,  1.])
        """
        if self.scalar == 0.0:
            raise ZeroDivisionError('{} not invertible'.format(self))

        return self.operator.inverse * (1.0 / self.scalar)

    def derivative(self, x):
        """Return the derivative at ``x``.

        Left scalar multiplication and derivative are commutative:

            ``OperatorLeftScalarMult(op, s).derivative(y) ==
            OperatorLeftScalarMult(op.derivative(y), s)``

        Parameters
        ----------
        x : `domain` `element-like`
            Evaluation point of the derivative

        Returns
        -------
        derivative : `OperatorLeftScalarMult`

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space) - space.element([1, 1, 1])
        >>> left_mul_op = OperatorLeftScalarMult(operator, 3)
        >>> derivative = left_mul_op.derivative([0, 0, 0])
        >>> derivative([1, 1, 1])
        rn(3).element([ 3.,  3.,  3.])
        """
        if self.is_linear:
            return self
        else:
            return self.scalar * self.operator.derivative(x)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator scalar multiplication is the
        scalar multiplication of the operator adjoint:

            ``OperatorLeftScalarMult(op, s).adjoint ==
            OperatorLeftScalarMult(op.adjoint, s)``

        Raises
        ------
        OpNotImplementedError
            If the underlying operator is non-linear.

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorLeftScalarMult(operator, 3)
        >>> left_mul_op.adjoint([1, 2, 3])
        rn(3).element([ 3.,  6.,  9.])
        """

        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        return self.scalar.conjugate() * self.operator.adjoint

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.scalar)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * {}'.format(self.scalar, self.operator)


class OperatorRightScalarMult(Operator):

    """Expression type for the operator right scalar multiplication.

        ``OperatorRightScalarMult(op, s) == op(s * x)``

    The scalar multiplication is well-defined only if ``op.domain`` is
    a `LinearSpace`.
    """

    def __init__(self, operator, scalar, tmp=None):
        """Initialize a new `OperatorLeftScalarMult` instance.

        Parameters
        ----------
        operator : `Operator`
            Operator in the scalar multiplication. Its `domain` must
            be a `LinearSpace` or `Field`.
        scalar : ``operator.range.field`` element
            A real or complex number, depending on the field of
            the operator domain.
        tmp : `domain` element, optional
            Used to avoid the creation of a temporary when applying the
            operator.

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorRightScalarMult(operator, 3)
        >>> left_mul_op([1, 2, 3])
        rn(3).element([ 3.,  6.,  9.])
        """
        if not isinstance(operator.domain, (LinearSpace, Field)):
            raise OpTypeError('domain {!r} not a `LinearSpace` or `Field` '
                              'instance'.format(operator.domain))

        if scalar not in operator.domain.field:
            raise TypeError('`scalar` {!r} not in the field {!r} of the '
                            'operator domain {!r}'
                            ''.format(scalar, operator.domain.field,
                                      operator.domain))

        if tmp is not None and tmp not in operator.domain:
            raise OpDomainError('`tmp` {!r} not an element of the '
                                'operator domain {!r}'
                                ''.format(tmp, operator.domain))

        if isinstance(operator, OperatorRightScalarMult):
            # Shortcut to save performance in case of repeated multiplications
            scalar = scalar * operator.scalar
            operator = operator.operator

        super(OperatorRightScalarMult, self).__init__(
            operator.domain, operator.range, operator.is_linear)
        self.__operator = operator
        self.__scalar = scalar
        self.__tmp = tmp

    @property
    def operator(self):
        """The operator part of this multiplication."""
        return self.__operator

    @property
    def scalar(self):
        """The scalar part of this multiplication."""
        return self.__scalar

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.operator(self.scalar * x)
        else:
            if self.__tmp is not None:
                tmp = self.__tmp
            else:
                tmp = self.domain.element()
            tmp.lincomb(self.scalar, x)
            self.operator(tmp, out=out)

    def __mul__(self, other):
        """Implement ``self * other``.

        An optimization for repeated multiplications.
        """

        if other in self.range.field:
            return OperatorRightScalarMult(self.operator, self.scalar * other,
                                           self.__tmp)
        else:
            return super(OperatorRightScalarMult, self).__rmul__(other)

    @property
    def inverse(self):
        """Inverse of this operator.

        The inverse of ``op * scalar`` is given by
        ``1/scalar * op.inverse`` if ``scalar != 0``. If ``scalar == 0``,
        the inverse is not defined.

            ``OperatorRightScalarMult(op, s).inverse ==
            OperatorLeftScalarMult(op.inverse, 1/s)``

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorRightScalarMult(operator, 3)
        >>> left_mul_op.inverse([3, 3, 3])
        rn(3).element([ 1.,  1.,  1.])
        """
        if self.scalar == 0.0:
            raise ZeroDivisionError('{} not invertible'.format(self))

        return (1.0 / self.scalar) * self.operator.inverse

    def derivative(self, x):
        """Return the derivative at ``x``.

        The derivative of the right scalar operator multiplication
        follows the chain rule:

            ``OperatorRightScalarMult(op, s).derivative(y) ==
            OperatorLeftScalarMult(op.derivative(s * y), s)``

        Parameters
        ----------
        x : `domain` `element-like`
            Evaluation point of the derivative.

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space) - space.element([1, 1, 1])
        >>> left_mul_op = OperatorRightScalarMult(operator, 3)
        >>> derivative = left_mul_op.derivative([0, 0, 0])
        >>> derivative([1, 1, 1])
        rn(3).element([ 3.,  3.,  3.])
        """
        return self.scalar * self.operator.derivative(self.scalar * x)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator scalar multiplication is the
        scalar multiplication of the operator adjoint:

        ``OperatorLeftScalarMult(op, s).adjoint ==
        OperatorLeftScalarMult(op.adjoint, s)``

        Raises
        ------
        OpNotImplementedError
            If the underlying operator is non-linear.

        Examples
        --------
        >>> space = odl.rn(3)
        >>> operator = odl.IdentityOperator(space)
        >>> left_mul_op = OperatorRightScalarMult(operator, 3)
        >>> left_mul_op.adjoint([1, 2, 3])
        rn(3).element([ 3.,  6.,  9.])
        """

        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        return self.operator.adjoint * self.scalar.conjugate()

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.scalar)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * {}'.format(self.operator, self.scalar)


class FunctionalLeftVectorMult(Operator):

    """Expression type for the functional left vector multiplication.

    A functional is an `Operator` whose `Operator.range` is
    a `Field`. It is multiplied from left with a `LinearSpaceElement`,
    resulting in an operator mapping from the `Operator.domain` to the
    element's `LinearSpaceElement.space`.

        ``FunctionalLeftVectorMult(op, y)(x) == y * op(x)``
    """

    def __init__(self, functional, vector):
        """Initialize a new instance.

        Parameters
        ----------
        functional : `Operator`
            Functional in the vector multiplication. Its `range` must
            be a `Field`.
        vector : ``functional.range`` `element-like`
            The element to multiply with. Its space's `LinearSpace.field`
            must be the same as ``functional.range``.

        Examples
        --------
        Create the operator ``(y * y^T)(x) = y * <x, y>``

        >>> space = odl.rn(3)
        >>> y = space.element([1, 2, 3])
        >>> functional = odl.InnerProductOperator(y)
        >>> left_mul_op = FunctionalLeftVectorMult(functional, y)
        >>> left_mul_op([1, 2, 3])
        rn(3).element([ 14.,  28.,  42.])
        """
        if not isinstance(vector, LinearSpaceElement):
            raise TypeError('`vector` {!r} not is not a LinearSpaceElement'
                            ''.format(vector))

        if functional.range != vector.space.field:
            raise OpTypeError('range {!r} not is not vector.space.field {!r}'
                              ''.format(functional.range, vector.space.field))

        super(FunctionalLeftVectorMult, self).__init__(
            functional.domain, vector.space, linear=functional.is_linear)
        self.__functional = functional
        self.__vector = vector

    @property
    def functional(self):
        """The functional part of this multiplication."""
        return self.__functional

    @property
    def vector(self):
        """The element part of this multiplication."""
        return self.__vector

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.vector * self.functional(x)
        else:
            scalar = self.functional(x)
            out.lincomb(scalar, self.vector)

    def derivative(self, x):
        """Return the derivative at ``x``.

        Left scalar multiplication and derivative are commutative:

            ``FunctionalLeftVectorMult(op, y).derivative(z) ==
            FunctionalLeftVectorMult(op.derivative(z), y)``

        Returns
        -------
        derivative : `FunctionalLeftVectorMult`
        """
        if self.is_linear:
            return self
        else:
            return FunctionalLeftVectorMult(self.functional.derivative(x),
                                            self.vector)

    @property
    def adjoint(self):
        """Adjoint of this operator.

            ``FunctionalLeftVectorMult(op, y).adjoint ==
            OperatorComp(op.adjoint, y.T)``

        Returns
        -------
        adjoint : `OperatorComp`

        Raises
        ------
        OpNotImplementedError
            If the underlying operator is non-linear.
        """

        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        return OperatorComp(self.functional.adjoint, self.vector.T)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.functional, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * {}'.format(self.vector, self.functional)


class OperatorLeftVectorMult(Operator):

    """Expression type for the operator left vector multiplication.

        ``OperatorLeftVectorMult(op, y)(x) <==> y * op(x)``
    """

    def __init__(self, operator, vector):
        """Initialize a new instance.

        Parameters
        ----------
        operator : `Operator`
            The range of ``op`` must be a `LinearSpace`.
        vector : `LinearSpaceElement` in ``op.range``
            The vector to multiply by
        """
        if vector not in operator.range:
            raise OpRangeError('`vector` {!r} not in operator.range {!r}'
                               ''.format(vector, operator.range))

        super(OperatorLeftVectorMult, self).__init__(
            operator.domain, operator.range, linear=operator.is_linear)
        self.__operator = operator
        self.__vector = vector

    @property
    def operator(self):
        """The operator part of this multiplication."""
        return self.__operator

    @property
    def vector(self):
        """The fixed element to multiply with."""
        return self.__vector

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.operator(x) * self.vector
        else:
            self.operator(x, out=out)
            out *= self.vector

    @property
    def inverse(self):
        """Inverse of this operator.

        The inverse of ``y * op`` is given by
        ``op.inverse / y``.

        ``OperatorLeftVectorMult(op, y).inverse ==
        OperatorRightVectorMult(op.inverse, 1/y)``
        """

        return self.operator.inverse * (1.0 / self.vector)

    def derivative(self, x):
        """Return the derivative at ``x``.

        Left scalar multiplication and derivative are commutative:

            ``OperatorLeftVectorMult(op, y).derivative(z) ==
            OperatorLeftVectorMult(op.derivative(z), y)``

        See Also
        --------
        OperatorLeftVectorMult : the result
        """
        if self.is_linear:
            return self
        else:
            return self.vector @ self.operator.derivative(x)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator left vector multiplication is the right
        multiplication of the given operator by the complex conjugate of the
        given vector.

            ``OperatorLeftVectorMult(op, y).adjoint ==
            OperatorRightVectorMult(op.adjoint, y.conj())``

        Returns
        -------
        adjoint : `OperatorRightVectorMult`

        Raises
        ------
        OpNotImplementedError
            If the underlying operator is non-linear.
        """
        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        if self.vector.space.is_real:
            # The complex conjugate of a real vector is the vector itself.
            return self.operator.adjoint * self.vector
        else:
            return self.operator.adjoint * self.vector.conj()

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * {}'.format(self.vector, self.operator)


class OperatorRightVectorMult(Operator):

    """Expression type for the operator right vector multiplication.

        ``OperatorRightVectorMult(op, y)(x) == op(y * x)``

    The scalar multiplication is well-defined only if ``y`` is in
    ``op.domain``.
    """

    def __init__(self, operator, vector):
        """Initialize a new `OperatorRightVectorMult` instance.

        Parameters
        ----------
        operator : `Operator`
            The domain of ``operator`` must be a ``vector.space``.
        vector : ``op.domain`` element
            The fixed element to multiply with.
        """
        if not isinstance(operator, Operator):
            raise TypeError('`operator` {!r} not an `Operator` instance'
                            ''.format(operator))

        if vector not in operator.domain:
            raise OpDomainError('`vector` {!r} not in operator.domain {!r}'
                                ''.format(vector.space, operator.domain))

        super(OperatorRightVectorMult, self).__init__(
            operator.domain, operator.range, linear=operator.is_linear)
        self.__operator = operator
        self.__vector = vector

    @property
    def operator(self):
        """The operator part of this multiplication."""
        return self.__operator

    @property
    def vector(self):
        """The fixed element to multiply with."""
        return self.__vector

    def _call(self, x, out=None):
        """Implement ``self(x[, out])``."""
        if out is None:
            return self.operator(x * self.vector)
        else:
            tmp = self.domain.element()
            x.multiply(self.vector, out=tmp)
            self.operator(tmp, out=out)

    @property
    def inverse(self):
        """Inverse of this operator.

        The inverse of ``op * y`` is given by
        ``(1/y) * op.inverse``.

            ``OperatorRightVectorMult(op, y).inverse ==
            OperatorLeftVectorMult(op.inverse, 1/y)``
        """
        return (1.0 / self.vector) * self.operator.inverse

    def derivative(self, x):
        """Return the derivative at ``x``.

        Left vector multiplication and derivative are commutative:

        ``OperatorRightVectorMult(op, y).derivative(x) <==>
        OperatorRightVectorMult(op.derivative(x), y)``

        See Also
        --------
        OperatorRightVectorMult : the result
        """
        if self.is_linear:
            return self
        else:
            return self.operator.derivative(self.vector * x) * self.vector

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint of the operator right vector multiplication is the left
        multiplication of the given operator by the complex conjugate of the
        given vector.

            ``OperatorRightVectorMult(op, y).adjoint ==
            OperatorLeftVectorMult(op.adjoint, y.conj())``

        Returns
        -------
        adjoint : `OperatorLeftVectorMult`

        Raises
        ------
        OpNotImplementedError
            If the underlying operator is non-linear.
        """
        if not self.is_linear:
            raise OpNotImplementedError('nonlinear operators have no adjoint')

        if self.vector.space.is_real:
            # The complex conjugate of a real vector is the vector itself.
            return self.vector @ self.operator.adjoint
        else:
            return self.vector.conj() @ self.operator.adjoint

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.operator, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * {}'.format(self.operator, self.vector)


class OpTypeError(TypeError):
    """Exception for operator type errors.

    Domain errors are raised by `Operator` subclasses when trying to call
    them with input not in the domain (`Operator.domain`) or with the wrong
    range (`Operator.range`).
    """


class OpDomainError(OpTypeError):
    """Exception for domain errors.

    Domain errors are raised by `Operator` subclasses when trying to call
    them with input not in the domain (`Operator.domain`).
    """


class OpRangeError(OpTypeError):
    """Exception for domain errors.

    Domain errors are raised by `Operator` subclasses when the returned
    value does not lie in the range (`Operator.range`).
    """


class OpNotImplementedError(NotImplementedError):
    """Exception for not implemented errors in `LinearSpace`'s.

    These are raised when a method in `LinearSpace` that has not been
    defined in a specific space is called.
    """


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Operators defined for tensor fields."""

from __future__ import absolute_import, division, print_function

from numbers import Integral
from typing import Optional

import numpy as np

from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY

from odl.core.operator.operator import Operator
from odl.core.set import ComplexNumbers, RealNumbers
from odl.core.space import ProductSpace, tensor_space
from odl.core.space.base_tensors import TensorSpace, Tensor
from odl.core.space.weightings.weighting import ArrayWeighting
from odl.core.util import dtype_repr, indent, signature_string
from odl.core.array_API_support import ArrayBackend, lookup_array_backend, abs as odl_abs, maximum, pow, sqrt, multiply, get_array_and_backend, can_cast, odl_all_equal

from odl.core.sparse import is_sparse, get_sparse_matrix_impl, lookup_sparse_format

__all__ = ('PointwiseNorm', 'PointwiseInner', 'PointwiseSum', 'MatrixOperator',
           'SamplingOperator', 'WeightedSumSamplingOperator',
           'FlatteningOperator', 'DeviceChangeOperator')

_SUPPORTED_DIFF_METHODS = ('central', 'forward', 'backward')


class DeviceChangeOperator(Operator):
    """An operator that is mathematically the identity, but whose domain and codomain
    differ in where they store their arrays.
    This is useful as an adaptor between operators that need to use different devices
    for some reason.
    Note that it is usually more efficient to implement your whole pipeline on a single
    device, if possible.
    """
    def __init__(self, domain=None, range=None, domain_device=None, range_device=None):
        """Create an operator tying two equivalent spaces with different storage together.

        Parameters
        ----------
        domain, range : `TensorSpace`, optional
            Spaces of vectors. Usually only one of them is specified; if both are
            given, they must be identical save for the device.
        domain_device, range_device : `str`, optional
            Device specifiers such as `'cpu'` or `'cuda:0'`. Which ones are
            supported depends on the backend and hardware.
            If e.g. `domain` and `range_device` are specified, the range will be
            chosen as `domain.to_device(range_device)`, vice versa.
        """
        if range is None:
            assert domain is not None
            assert range_device is not None
            assert domain_device is None or domain_device == domain.device
            range = domain.to_device(range_device)
        elif domain is None:
            assert range is not None
            assert domain_device is not None
            assert range_device is None or range_device == range.device
            domain = range.to_device(domain_device)
        else:
            assert domain_device is None or domain_device == domain.device
            assert range_device is None or range_device == range.device
            assert(domain.to_device(range.device) == range)
        super().__init__(domain, range=range, linear=True)

    def _call(self, x):
        """Copy data to the intended device."""
        return x.to_device(self.range.device)

    @property
    def inverse(self):
        """Operator that copies data back to the original device."""
        return DeviceChangeOperator(domain=self.range, range=self.domain)

    @property
    def adjoint(self):
        """Adjoint is the same as inverse, as device change is mathematically
        the identity."""
        return self.inverse

    def norm(self, estimate=False, **kwargs):
        """Return the operator norm of this operator. This is 1, as the
        operator is mathematically the identity."""
        return 1

    def __repr__(self):
        """Represent the operator by its domain and the device of the range."""
        return f"{self.__class__.__name__}(domain={repr(self.domain)}, range_device={repr(self.range.device)})"

    def __str__(self):
        return f"{self.__class__.__name__}(domain={str(self.domain)}, range_device={str(self.range.device)})"

class ImplChangeOperator(Operator):
    """An operator that is mathematically the identity, but whose domain and codomain
    differ in what backend they use for their arrays.
    This is useful as an adaptor between operators that need to use different backend
    for some reason, for example one operator implemented through bespoke C code and
    one operator implemented with PyTorch neural networks.
    Note that it is usually more efficient to keep your whole pipeline on a single
    backend and device, if possible.
    """
    def __init__(self, domain=None, range=None, domain_impl=None, range_impl=None):
        """Create an operator tying two equivalent spaces with different storage together.

        Parameters
        ----------
        domain, range : `TensorSpace`, optional
            Spaces of vectors. Usually only one of them is specified; if both are
            given, they must be identical save for the backend (`impl`).
        domain_impl, range_impl : `str`, optional
            Backend identifier. Must correspond to a registered backend,
            cf. `odl.core.space.entry_points.tensor_space_impl_names`.
            If e.g. `domain` and `range_impl` are specified, the range will be
            chosen as `domain.to_impl(range_impl)`, vice versa.
            The device of the space must be usable simultaneously with both of
            the backends.
        """
        if range is None:
            assert impl is not None
            assert range_impl is not None
            assert domain_impl is None or domain_impl == domain.impl
            range = domain.to_impl(range_impl)
        elif domain is None:
            assert range is not None
            assert domain_impl is not None
            assert range_impl is None or range_impl == range.impl
            domain = range.to_impl(domain_impl)
        else:
            assert domain_impl is None or domain_impl == domain.impl
            assert range_impl is None or range_impl == range.impl
            assert(domain.to_impl(range.impl) == range)
        super().__init__(domain, range=range, linear=True)

    def _call(self, x):
        """Copy data to the intended backend."""
        return x.to_impl(self.range.impl)

    @property
    def inverse(self):
        """Operator that copies data back to the original backend."""
        return ImplChangeOperator(domain=self.range, range=self.domain)

    @property
    def adjoint(self):
        """Adjoint is the same as inverse, as backend change is mathematically
        the identity."""
        return self.inverse

    def norm(self, estimate=False, **kwargs):
        """Return the operator norm of this operator. This is 1, as the
        operator is mathematically the identity."""
        return 1

    def __repr__(self):
        """Represent the operator by its domain and the impl of the range."""
        return f"{self.__class__.__name__}(domain={repr(self.domain)}, range_impl={repr(self.range.impl)})"

    def __str__(self):
        return f"{self.__class__.__name__}(domain={str(self.domain)}, range_impl={str(self.range.impl)})"

class PointwiseTensorFieldOperator(Operator):

    """Abstract operator for point-wise tensor field manipulations.

    A point-wise operator acts on a space of vector or tensor fields,
    i.e. a power space ``X^d`` of a discretized function space ``X``.
    Its range is the power space ``X^k`` with a possibly different
    number ``k`` of components. For ``k == 1``, the base space
    ``X`` can be used instead.

    For example, if ``X`` is a `DiscretizedSpace` space, then
    ``ProductSpace(X, d)`` is a valid domain for any positive integer
    ``d``. It is also possible to have tensor fields over tensor fields, i.e.
    ``ProductSpace(ProductSpace(X, n), m)``.

    .. note::
        It is allowed that ``domain``, ``range`` and ``base_space`` use
        different ``dtype``. Correctness for, e.g., real-to-complex mappings
        is not guaranteed in that case.

    See Also
    --------
    odl.core.space.pspace.ProductSpace
    """

    def __init__(self, domain, range, base_space, linear=False):
        """Initialize a new instance.

        Parameters
        ----------
        domain, range : {`ProductSpace`, `LinearSpace`}
            Spaces of vector fields between which the operator maps.
            They have to be either power spaces of the same base space
            ``X`` (up to ``dtype``), or the base space itself.
            Empty product spaces are not allowed.
        base_space : `LinearSpace`
            The base space ``X``.
        linear : bool, optional
            If ``True``, assume that the operator is linear.
        """
        if not is_compatible_space(domain, base_space):
            raise ValueError(
                '`domain` {!r} is not compatible with `base_space` {!r}'
                ''.format(domain, base_space))

        if not is_compatible_space(range, base_space):
            raise ValueError(
                '`range` {!r} is not compatible with `base_space` {!r}'
                ''.format(range, base_space))

        super(PointwiseTensorFieldOperator, self).__init__(
            domain=domain, range=range, linear=linear)
        self.__base_space = base_space

    @property
    def base_space(self):
        """Base space ``X`` of this operator's domain and range."""
        return self.__base_space


class PointwiseNorm(PointwiseTensorFieldOperator):

    """Take the point-wise norm of a vector field.

    This operator computes the (weighted) p-norm in each point of
    a vector field, thus producing a scalar-valued function.
    It implements the formulas ::

        ||F(x)|| = [ sum_j( w_j * |F_j(x)|^p ) ]^(1/p)

    for ``p`` finite and ::

        ||F(x)|| = max_j( w_j * |F_j(x)| )

    for ``p = inf``, where ``F`` is a vector field. This implies that
    the `Operator.domain` is a power space of a discretized function
    space. For example, if ``X`` is a `DiscretizedSpace` space, then
    ``ProductSpace(X, d)`` is a valid domain for any positive integer
    ``d``.
    """

    def __init__(self, vfspace, exponent=None, weighting=None):
        """Initialize a new instance.

        Parameters
        ----------
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        exponent : non-zero float, optional
            Exponent of the norm in each point. Values between
            0 and 1 are currently not supported due to numerical
            instability.
            Default: ``vfspace.exponent``
        weighting : `array-like` or positive float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``len(domain)``, and
            all entries must be positive.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.

        Examples
        --------
        We make a tiny vector field space in 2D and create the
        standard point-wise norm operator on that space. The operator
        maps a vector field to a scalar function:

        >>> spc = odl.uniform_discr([-1, -1], [1, 1], (1, 2))
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> pw_norm = odl.PointwiseNorm(vfspace)
        >>> pw_norm.range == spc
        True

        Now we can calculate the 2-norm in each point:

        >>> x = vfspace.element([[[1, -4]],
        ...                      [[0, 3]]])
        >>> print(pw_norm(x))
        [[ 1.,  5.]]

        We can change the exponent either in the vector field space
        or in the operator directly:

        >>> vfspace = odl.ProductSpace(spc, 2, exponent=1)
        >>> pw_norm = PointwiseNorm(vfspace)
        >>> print(pw_norm(x))
        [[ 1.,  7.]]
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> pw_norm = PointwiseNorm(vfspace, exponent=1)
        >>> print(pw_norm(x))
        [[ 1.,  7.]]
        """
        if not isinstance(vfspace, ProductSpace):
            raise TypeError('`vfspace` {!r} is not a ProductSpace '
                            'instance'.format(vfspace))
        super(PointwiseNorm, self).__init__(
            domain=vfspace, range=vfspace[0].real_space, base_space=vfspace[0],
            linear=False)

        # Need to check for product space shape once higher order tensors
        # are implemented

        if exponent is None:
            if self.domain.exponent is None:
                raise ValueError('cannot determine `exponent` from {}'
                                 ''.format(self.domain))
            self._exponent = self.domain.exponent
        elif exponent < 1:
            raise ValueError('`exponent` smaller than 1 not allowed')
        else:
            self._exponent = float(exponent)

        # Handle weighting, including sanity checks
        if weighting is None:
            # TODO: find a more robust way of getting the weights as an array
            if hasattr(self.domain.weighting, 'array'):
                self.__weights = self.domain.weighting.array
            elif hasattr(self.domain.weighting, 'const'):
                self.__weights = [self.domain.weighting.const *self.domain[i].one() for i in range(len(vfspace))]
            else:
                raise ValueError('weighting scheme {!r} of the domain does '
                                 'not define a weighting array or constant'
                                 ''.format(self.domain.weighting))
            self.__is_weighted = False

        else:
            ### This is a bad situation: although we worked hard to get an elegant weighting, the PointwiseNorm just yanks all of that down the drain by reimplementing the norm operation and the input sanitisation just for a ProductSpace. 
            ### EV reimplemented these two functionnalities but moving forward, this should be coerced into abiding to our new API

            if isinstance(weighting, list) and all([isinstance(w, Tensor) for w in weighting]) :
                self.__weights = weighting
                self.__is_weighted = all([odl_all_equal(w, 1) for w in weighting])
            else:
                if isinstance(weighting, (int, float)):
                    weighting = [weighting for _ in range(len(self.domain))]

                weighted_flag = []
                for i in range(len(self.domain)):
                    if weighting[i] <= 0:
                        raise ValueError(f'weighting array weighting contains invalid entry {weighting[i]}')
                    if weighting[i] in [1,1.0]:
                        weighted_flag.append(False)
                    else:
                        weighted_flag.append(True)
                self.__is_weighted = True if any(weighted_flag) else False

                weighting = [
                    self.domain[i].tspace.broadcast_to(weighting[i]) 
                    for i in range(len(self.domain))
                    ]
                
                self.__weights = []
                for i in range(len(self.domain)):
                    self.__weights.append(self.domain[i].element(weighting[i]))       

    @property
    def exponent(self):
        """Exponent ``p`` of this norm."""
        return self._exponent

    @property
    def weights(self):
        """Weighting array of this operator."""
        return self.__weights

    @property
    def is_weighted(self):
        """``True`` if weighting is not 1 or all ones."""
        return self.__is_weighted

    def _call(self, f, out):
        """Implement ``self(f, out)``."""
        if self.exponent == 1.0:
            self._call_vecfield_1(f, out)
        elif self.exponent == float('inf'):
            self._call_vecfield_inf(f, out)
        else:
            self._call_vecfield_p(f, out)

    def _call_vecfield_1(self, vf, out):
        """Implement ``self(vf, out)`` for exponent 1."""
        
        odl_abs(vf[0], out=out)
        if self.is_weighted:
            out *= self.weights[0]

        if len(self.domain) == 1:
            return

        tmp = self.range.element()
        for fi, wi in zip(vf[1:], self.weights[1:]):
            odl_abs(fi, out=tmp)
            if self.is_weighted:
                tmp *= wi
            out += tmp

    def _call_vecfield_inf(self, vf, out):
        """Implement ``self(vf, out)`` for exponent ``inf``."""
        odl_abs(vf[0], out=out)
        if self.is_weighted:
            out *= self.weights[0]

        if len(self.domain) == 1:
            return

        tmp = self.range.element()
        for vfi, wi in zip(vf[1:], self.weights[1:]):
            odl_abs(vfi, out=tmp)
            if self.is_weighted:
                tmp *= wi
            maximum(out, tmp, out=out)

    def _call_vecfield_p(self, vf, out):
        """Implement ``self(vf, out)`` for exponent 1 < p < ``inf``."""
        # Optimization for 1 component - just absolute value (maybe weighted)
        if len(self.domain) == 1:
            odl_abs(vf[0], out=out)
            if self.is_weighted:
                out *= self.weights[0] ** (1 / self.exponent)
            return

        # Initialize out, avoiding one copy
        self._abs_pow(vf[0], out=out, p=self.exponent)
        if self.is_weighted:
            out *= self.weights[0]

        tmp = self.range.element()
        for fi, wi in zip(vf[1:], self.weights[1:]):
            self._abs_pow(fi, out=tmp, p=self.exponent)
            if self.is_weighted:
                tmp *= wi
            out += tmp

        self._abs_pow(out, out=out, p=(1 / self.exponent))

    def _abs_pow(self, fi, out, p):
        """Compute |F_i(x)|^p point-wise and write to ``out``."""
        # Optimization for very common cases        
        if p == 0.5:
            odl_abs(fi, out=out)
            sqrt(out, out=out)
        elif p == 2.0 and self.base_space.field == RealNumbers():
            multiply(fi, fi, out=out)
        else:
            odl_abs(fi, out=out)
            pow(out, p, out=out)

    def derivative(self, vf):
        """Derivative of the point-wise norm operator at ``vf``.

        The derivative at ``F`` of the point-wise norm operator ``N``
        with finite exponent ``p`` and weights ``w`` is the pointwise
        inner product with the vector field ::

            x --> N(F)(x)^(1-p) * [ F_j(x) * |F_j(x)|^(p-2) ]_j

        Note that this is not well-defined for ``F = 0``. If ``p < 2``,
        any zero component will result in a singularity.

        Parameters
        ----------
        vf : `domain` `element-like`
            Vector field ``F`` at which to evaluate the derivative.

        Returns
        -------
        deriv : `PointwiseInner`
            Derivative operator at the given point ``vf``.

        Raises
        ------
        NotImplementedError
            * if the vector field space is complex, since the derivative
              is not linear in that case
            * if the exponent is ``inf``
        """
        if self.domain.field == ComplexNumbers():
            raise NotImplementedError('operator not Frechet-differentiable '
                                      'on a complex space')

        if self.exponent == float('inf'):
            raise NotImplementedError('operator not Frechet-differentiable '
                                      'for exponent = inf')

        vf = self.domain.element(vf)
        vf_pwnorm_fac = self(vf)
        if self.exponent != 2:  # optimize away most common case.
            vf_pwnorm_fac **= (self.exponent - 1)

        inner_vf = vf.copy()

        for gi in inner_vf:
            gi *= pow(odl_abs(gi), self.exponent - 2)
            if self.exponent >= 2:
                # Any component that is zero is not divided with
                nz = (vf_pwnorm_fac.asarray() != 0)
                gi[nz] /= vf_pwnorm_fac[nz]
            else:
                # For exponents < 2 there will be a singularity if any
                # component is zero. This results in inf or nan. See the
                # documentation for further details.
                gi /= vf_pwnorm_fac

        return PointwiseInner(self.domain, inner_vf, weighting=self.weights)


class PointwiseInnerBase(PointwiseTensorFieldOperator):
    """Base class for `PointwiseInner` and `PointwiseInnerAdjoint`.

    Implemented to allow code reuse between the classes.
    """

    def __init__(self, adjoint, vfspace, vecfield, weighting=None):
        """Initialize a new instance.

        All parameters are given according to the specifics of the "usual"
        operator. The ``adjoint`` parameter is used to control conversions
        for the inverse transform.

        Parameters
        ----------
        adjoint : bool
            ``True`` if the operator should be the adjoint, ``False``
            otherwise.
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        vecfield : ``vfspace`` `element-like`
            Vector field with which to calculate the point-wise inner
            product of an input vector field
        weighting : `array-like` or float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``len(domain)``.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.
        """
        if not isinstance(vfspace, ProductSpace):
            raise TypeError('`vfspace` {!r} is not a ProductSpace '
                            'instance'.format(vfspace))
        if adjoint:
            super(PointwiseInnerBase, self).__init__(
                domain=vfspace[0], range=vfspace, base_space=vfspace[0],
                linear=True)
        else:
            super(PointwiseInnerBase, self).__init__(
                domain=vfspace, range=vfspace[0], base_space=vfspace[0],
                linear=True)

        # Bail out if the space is complex but we cannot take the complex
        # conjugate.
        if (vfspace.field == ComplexNumbers() and
                not hasattr(self.base_space.element_type, 'conj')):
            raise NotImplementedError(
                'base space element type {!r} does not implement conj() '
                'method required for complex inner products'
                ''.format(self.base_space.element_type))

        self._vecfield = vfspace.element(vecfield)

        # Handle weighting, including sanity checks
        if weighting is None:
            self.__is_weighted =  False
            if hasattr(vfspace.weighting, 'array'):
                self.__weights = vfspace.weighting.array
            elif hasattr(vfspace.weighting, 'const'):
                # Casting the constant to an array of constants is just bad
                self.__weights = [vfspace.weighting.const *vfspace[i].one() for i in range(len(vfspace))]
            else:
                raise ValueError('weighting scheme {!r} of the domain does '
                                 'not define a weighting array or constant'
                                 ''.format(vfspace.weighting))
            
        else:
            # Check if the input has already been sanitised, i.e is it an odl.Tensor
            if isinstance(weighting, list) and all([isinstance(w, Tensor) for w in weighting]) :
                self.__weights = weighting
                self.__is_weighted = all([odl_all_equal(w, 1) for w in weighting])

            # these are required to provide an array-API compatible weighting parsing.
            else:
                if isinstance(weighting, (int, float)):
                    weighting = [weighting for i in range(len(vfspace))]

                weighted_flag = []
                for i in range(len(vfspace)):
                    if weighting[i] <= 0:
                        raise ValueError(f'weighting array weighting contains invalid entry {weighting[i]}')
                    if weighting[i] in [1,1.0]:
                        weighted_flag.append(False)
                    else:
                        weighted_flag.append(True)
                self.__is_weighted = True if any(weighted_flag) else False

                weighting = [
                    vfspace[i].tspace.broadcast_to(weighting[i]) 
                    for i in range(len(vfspace))
                    ]
                
                self.__weights = []
                for i in range(len(vfspace)):
                    self.weights.append(vfspace[i].element(weighting[i]))

    @property
    def vecfield(self):
        """Fixed vector field ``G`` of this inner product."""
        return self._vecfield

    @property
    def weights(self):
        """Weighting array of this operator."""
        return self.__weights

    @property
    def is_weighted(self):
        """``True`` if weighting is not 1 or all ones."""
        return self.__is_weighted

    @property
    def adjoint(self):
        """Adjoint operator."""
        raise NotImplementedError('abstract method')


class PointwiseInner(PointwiseInnerBase):

    """Take the point-wise inner product with a given vector field.

    This operator takes the (weighted) inner product ::

        <F(x), G(x)> = sum_j ( w_j * F_j(x) * conj(G_j(x)) )

    for a given vector field ``G``, where ``F`` is the vector field
    acting as a variable to this operator.

    This implies that the `Operator.domain` is a power space of a
    discretized function space. For example, if ``X`` is a `DiscretizedSpace`
    space, then ``ProductSpace(X, d)`` is a valid domain for any
    positive integer ``d``.
    """

    def __init__(self, vfspace, vecfield, weighting=None):
        """Initialize a new instance.

        Parameters
        ----------
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        vecfield : ``vfspace`` `element-like`
            Vector field with which to calculate the point-wise inner
            product of an input vector field
        weighting : `array-like` or float, optional
            Weighting array or constant for the norm. If an array is
            given, its length must be equal to ``len(domain)``, and
            all entries must be positive.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.

        Examples
        --------
        We make a tiny vector field space in 2D and create the
        point-wise inner product operator with a fixed vector field.
        The operator maps a vector field to a scalar function:

        >>> spc = odl.uniform_discr([-1, -1], [1, 1], (1, 2))
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> fixed_vf = np.array([[[0, 1]],
        ...                      [[1, -1]]])
        >>> pw_inner = PointwiseInner(vfspace, fixed_vf)
        >>> pw_inner.range == spc
        True

        Now we can calculate the inner product in each point:

        >>> x = vfspace.element([[[1, -4]],
        ...                      [[0, 3]]])
        >>> print(pw_inner(x))
        [[ 0., -7.]]
        """
        super(PointwiseInner, self).__init__(
            adjoint=False, vfspace=vfspace, vecfield=vecfield,
            weighting=weighting)

    @property
    def vecfield(self):
        """Fixed vector field ``G`` of this inner product."""
        return self._vecfield

    def _call(self, vf, out):
        """Implement ``self(vf, out)``."""
        if self.domain.field == ComplexNumbers():
            vf[0].multiply(self._vecfield[0].conj(), out=out)
        else:
            vf[0].multiply(self._vecfield[0], out=out)

        if self.is_weighted:
            out *= self.weights[0]

        if len(self.domain) == 1:
            return

        tmp = self.range.element()
        for vfi, gi, wi in zip(vf[1:], self.vecfield[1:],
                               self.weights[1:]):

            if self.domain.field == ComplexNumbers():
                vfi.multiply(gi.conj(), out=tmp)
            else:
                vfi.multiply(gi, out=tmp)

            if self.is_weighted:
                tmp *= wi
            out += tmp

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `PointwiseInnerAdjoint`
        """
        return PointwiseInnerAdjoint(
            sspace=self.base_space, vecfield=self.vecfield,
            vfspace=self.domain, weighting=self.weights)


class PointwiseInnerAdjoint(PointwiseInnerBase):

    """Adjoint of the point-wise inner product operator.

    The adjoint of the inner product operator is a mapping ::

        A^* : X --> X^d

    If the vector field space ``X^d`` is weighted by a vector ``v``,
    the adjoint, applied to a function ``h`` from ``X`` is the vector
    field ::

        x --> h(x) * (w / v) * G(x),

    where ``G`` and ``w`` are the vector field and weighting from the
    inner product operator, resp., and all multiplications are understood
    component-wise.
    """

    def __init__(self, sspace, vecfield, vfspace=None, weighting=None):
        """Initialize a new instance.

        Parameters
        ----------
        sspace : `LinearSpace`
            "Scalar" space on which the operator acts
        vecfield : range `element-like`
            Vector field of the point-wise inner product operator
        vfspace : `ProductSpace`, optional
            Space of vector fields to which the operator maps. It must
            be a power space with ``sspace`` as base space.
            This option is intended to enforce an operator range
            with a certain weighting.
            Default: ``ProductSpace(space, len(vecfield),
            weighting=weighting)``
        weighting : `array-like` or float, optional
            Weighting array or constant of the inner product operator.
            If an array is given, its length must be equal to
            ``len(vecfield)``.
            By default, the weights are is taken from
            ``range.weighting`` if applicable. Note that this excludes
            unusual weightings with custom inner product, norm or dist.
        """
        if vfspace is None:
            vfspace = ProductSpace(sspace, len(vecfield), weighting=weighting)
        else:
            if not isinstance(vfspace, ProductSpace):
                raise TypeError('`vfspace` {!r} is not a '
                                'ProductSpace instance'.format(vfspace))
            if vfspace[0] != sspace:
                raise ValueError('base space of the range is different from '
                                 'the given scalar space ({!r} != {!r})'
                                 ''.format(vfspace[0], sspace))
        super(PointwiseInnerAdjoint, self).__init__(
            adjoint=True, vfspace=vfspace, vecfield=vecfield,
            weighting=weighting)

        # Get weighting from range
        if hasattr(self.range.weighting, 'array'):
            ### The tolist() is an ugly tweak to recover a list from the pspace weighting.array which is stored in numpy 
            self.__ran_weights = vfspace.element(self.range.weighting.array.tolist())
        elif hasattr(self.range.weighting, 'const'):
            # Casting the constant to an array of constants is just bad
            self.__ran_weights = [self.range.weighting.const *self.range[i].one() for i in range(len(self.range))]
            
        else:
            raise ValueError('weighting scheme {!r} of the range does '
                             'not define a weighting array or constant'
                             ''.format(self.range.weighting))

    def _call(self, f, out):
        """Implement ``self(vf, out)``."""
        for vfi, oi, ran_wi, dom_wi in zip(self.vecfield, out,
                                           self.__ran_weights, self.weights):
            vfi.multiply(f, out=oi)
            # Removed the optimisation here, it would require casting ran_wi as odl.TensorSpaceElement
            # if not isclose(ran_wi, dom_wi).all():
            oi *= dom_wi / ran_wi

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `PointwiseInner`
        """
        return PointwiseInner(vfspace=self.range, vecfield=self.vecfield,
                              weighting=self.weights)


# TODO: Make this an optimized operator on its own.
class PointwiseSum(PointwiseInner):

    """Take the point-wise sum of a vector field.

    This operator takes the (weighted) sum ::

        sum(F(x)) = [ sum_j( w_j * F_j(x) ) ]

    where ``F`` is a vector field. This implies that
    the `Operator.domain` is a power space of a discretized function
    space. For example, if ``X`` is a `DiscretizedSpace` space, then
    ``ProductSpace(X, d)`` is a valid domain for any positive integer
    ``d``.
    """

    def __init__(self, vfspace, weighting=None):
        """Initialize a new instance.

        Parameters
        ----------
        vfspace : `ProductSpace`
            Space of vector fields on which the operator acts.
            It has to be a product space of identical spaces, i.e. a
            power space.
        weighting : `array-like` or float, optional
            Weighting array or constant for the sum. If an array is
            given, its length must be equal to ``len(domain)``.
            By default, the weights are is taken from
            ``domain.weighting``. Note that this excludes unusual
            weightings with custom inner product, norm or dist.

        Examples
        --------
        We make a tiny vector field space in 2D and create the
        standard point-wise sum operator on that space. The operator
        maps a vector field to a scalar function:

        >>> spc = odl.uniform_discr([-1, -1], [1, 1], (1, 2))
        >>> vfspace = odl.ProductSpace(spc, 2)
        >>> pw_sum = PointwiseSum(vfspace)
        >>> pw_sum.range == spc
        True

        Now we can calculate the sum in each point:

        >>> x = vfspace.element([[[1, -4]],
        ...                      [[0, 3]]])
        >>> print(pw_sum(x))
        [[ 1., -1.]]
        """
        if not isinstance(vfspace, ProductSpace):
            raise TypeError('`vfspace` {!r} is not a ProductSpace '
                            'instance'.format(vfspace))

        ones = vfspace.one()
        super(PointwiseSum, self).__init__(
            vfspace, vecfield=ones, weighting=weighting)


class MatrixOperator(Operator):

    """A matrix acting as a linear operator.

    This operator uses a matrix to represent an operator, and get its
    adjoint and inverse by doing computations on the matrix. This is in
    general a rather slow and memory-inefficient approach, and users are
    recommended to use other alternatives if possible.
    """

    def __init__(self, matrix, domain=None, range=None,
                 impl: Optional[str]=None,
                 device: Optional[str]=None,
                 axis=0):
        r"""Initialize a new instance.

        Parameters
        ----------
        matrix : `array-like` or `scipy.sparse.base.spmatrix`
            2-dimensional array representing the linear operator.
            For Scipy sparse matrices only tensor spaces with
            ``ndim == 1`` are allowed as ``domain``.
            The matrix is copied to `impl`/`device`, if these are
            specified (only once, when the operator is initialized).
            If a plain Python list is supplied, it will first
            be converted to a NumPy array.
        domain : `TensorSpace`, optional
            Space of elements on which the operator can act. Its
            ``dtype`` must be castable to ``range.dtype``.
            For the default ``None``, a space with 1 axis and size
            ``matrix.shape[1]`` is used, together with the matrix'
            data type.
        range : `TensorSpace`, optional
            Space of elements on to which the operator maps. Its
            ``shape`` and ``dtype`` attributes must match those
            of the result of the multiplication.
            For the default ``None``, the range is inferred from
            ``matrix``, ``domain`` and ``axis``.
        impl : `ArrayBackend`-identifying `str`, optional
            Which backend to use for the low-level matrix multiplication.
            If not explicitly provided, it will be inferred in the following
            order of preference, depending on what is available:
            1. from `domain`
            2. from `range`
            3. from `matrix`
        device : `str`, optional
            On which device to store the matrix.
            Same defaulting logic as for `impl`.
        axis : int, optional
            Sum over this axis of an input tensor in the
            multiplication.

        Examples
        --------
        By default, ``domain`` and ``range`` are spaces of with one axis:

        >>> m = np.ones((3, 4))
        >>> op = MatrixOperator(m)
        >>> op.domain
        rn(4)
        >>> op.range
        rn(3)
        >>> op([1, 2, 3, 4])
        rn(3).element([ 10.,  10.,  10.])

        For multi-dimensional arrays (tensors), the summation
        (contraction) can be performed along a specific axis. In
        this example, the number of matrix rows (4) must match the
        domain shape entry in the given axis:

        >>> dom = odl.rn((5, 4, 4))  # can use axis=1 or axis=2
        >>> op = MatrixOperator(m, domain=dom, axis=1)
        >>> op(dom.one()).shape
        (5, 3, 4)
        >>> op = MatrixOperator(m, domain=dom, axis=2)
        >>> op(dom.one()).shape
        (5, 4, 3)

        The operator also works on `uniform_discr` type spaces. Note,
        however, that the ``weighting`` of the domain is propagated to
        the range by default, in order to keep the correspondence between
        adjoint and transposed matrix:

        >>> space = odl.uniform_discr(0, 1, 4)
        >>> op = MatrixOperator(m, domain=space)
        >>> op(space.one())
        rn(3, weighting=0.25).element([ 4.,  4.,  4.])
        >>> np.array_equal(op.adjoint.matrix, m.T)
        True

        Notes
        -----
        For a matrix :math:`A \in \mathbb{F}^{n \times m}`, the
        operation on a tensor :math:`T \in \mathbb{F}^{n_1 \times
        \dots \times n_d}` is defined as the summation

        .. math::
            (A \cdot T)_{i_1, \dots, i_k, \dots, i_d} =
            \sum_{j=1}^m A_{i_k j} T_{i_1, \dots, j, \dots, i_d}.

        It produces a new tensor :math:`A \cdot T \in \mathbb{F}^{
        n_1 \times \dots \times n \times \dots \times n_d}`.
        """
        def infer_backend_from(default_backend):
            if impl is not None:
                self.__array_backend = lookup_array_backend(impl)
            else:
                assert(isinstance(default_backend, ArrayBackend))
                self.__array_backend = default_backend
        def infer_device_from(default_device):
            self.__device = default_device if device is None else device

        self._sparse_format = lookup_sparse_format(matrix)

        if domain is not None:
            infer_backend_from(domain.array_backend)
            infer_device_from(domain.device)
            
        elif range is not None:
            infer_backend_from(range.array_backend)
            infer_device_from(range.device)

        elif self.is_sparse:
            if self._sparse_format.impl == 'scipy':
                infer_backend_from(lookup_array_backend('numpy'))
                infer_device_from('cpu')

            elif self._sparse_format.impl == 'pytorch':
                infer_backend_from(lookup_array_backend('pytorch'))
                infer_device_from(matrix.device)
                
            else:
                raise ValueError
        
        elif isinstance(matrix, (list, tuple)):
            infer_backend_from(lookup_array_backend('numpy'))
            infer_device_from('cpu')
        else:
            infer_backend_from(get_array_and_backend(matrix)[1])
            infer_device_from(matrix.device)

        self.__arr_ns = self.array_backend.array_namespace

        if self.is_sparse:
            if self._sparse_format.impl == 'scipy':
                if self.array_backend.impl != 'numpy':
                    raise TypeError(f"SciPy sparse matrices can only be used with NumPy on CPU, not {self.array_backend.impl}.")
                if self.device != 'cpu':
                    raise TypeError(f"SciPy sparse matrices can only be used with NumPy on CPU, not {device}.")
            elif self._sparse_format.impl == 'pytorch':
                if self.array_backend.impl != 'pytorch':
                    raise TypeError(f"PyTorch sparse matrices can only be used with Pytorch, not {self.array_backend.impl}.")
            self.__matrix = matrix

        elif isinstance(matrix, Tensor):
            self.__matrix = matrix.data
            self.__matrix = self.__arr_ns.asarray(matrix.data, device=self.__device, copy=AVOID_UNNECESSARY_COPY)
            while len(self.__matrix.shape) < 2:
                self.__matrix = self.__matrix[None]
        else:
            self.__matrix = self.__arr_ns.asarray(matrix, device=self.__device, copy=AVOID_UNNECESSARY_COPY)
            while len(self.__matrix.shape) < 2:
                self.__matrix = self.__matrix[None]

        self.__axis, axis_in = int(axis), axis
        if self.axis != axis_in:
            raise ValueError('`axis` must be integer, got {}'.format(axis_in))

        if self.matrix.ndim != 2:
            raise ValueError('`matrix` has {} axes instead of 2'
                             ''.format(self.matrix.ndim))

        # Infer or check domain
        if domain is None:
            dtype = self.array_backend.identifier_of_dtype(self.matrix.dtype)
            domain = tensor_space((self.matrix.shape[1],),
                                  dtype=dtype,
                                  impl = self.array_backend.impl,
                                  device = self.device
                                  )
        else:
            if not isinstance(domain, TensorSpace):
                raise TypeError('`domain` must be a `TensorSpace` '
                                'instance, got {!r}'.format(domain))

            if self.is_sparse and domain.ndim > 1:
                raise ValueError('`domain.ndim` > 1 unsupported for '
                                 'scipy sparse matrices')

            if domain.shape[axis] != self.matrix.shape[1]:
                raise ValueError('`domain.shape[axis]` not equal to '
                                 '`matrix.shape[1]` ({} != {})'
                                 ''.format(domain.shape[axis],
                                           self.matrix.shape[1]))

        range_shape = list(domain.shape)
        range_shape[self.axis] = self.matrix.shape[0]

        if range is None:
            # Infer range
            range_dtype = self.__arr_ns.result_type(
                self.matrix.dtype, domain.dtype)
            range_dtype = self.array_backend.identifier_of_dtype(range_dtype)
            if (range_shape != domain.shape and
                    isinstance(domain.weighting, ArrayWeighting)):
                # Cannot propagate weighting due to size mismatch.
                weighting = None
            else:
                weighting = domain.weighting
            range = tensor_space(range_shape, 
                                 impl = self.array_backend.impl,
                                 device=self.device,
                                 dtype=range_dtype,
                                 weighting=weighting,
                                 exponent=domain.exponent)
        else:
            # Check consistency of range
            if not isinstance(range, TensorSpace):
                raise TypeError('`range` must be a `TensorSpace` instance, '
                                'got {!r}'.format(range))

            if range.shape != tuple(range_shape):
                raise ValueError('expected `range.shape` = {}, got {}'
                                 ''.format(tuple(range_shape), range.shape))

        # Check compatibility of data types
        result_dtype = self.__arr_ns.result_type(domain.dtype, self.matrix.dtype)
        if not can_cast(self.__arr_ns, result_dtype, range.dtype):
            raise ValueError('result data type {} cannot be safely cast to '
                             'range data type {}'
                             ''.format(dtype_repr(result_dtype),
                                       dtype_repr(range.dtype)))

        super(MatrixOperator, self).__init__(domain, range, linear=True)

    @property
    def is_sparse(self):
        return self._sparse_format is not None

    @property
    def matrix(self):
        """Matrix representing this operator."""
        return self.__matrix

    @property
    def array_backend(self):
        """Backend on which to carry out the BLAS matmul operation.
        Note that this does not necessarily have to be the same as
        either the range or domain of the operator, but by default it will
        be chosen such. If a different backend and/or device is used, the
        operator will always copy data to `self.array_backend` before
        carrying out the matrix multiplication, then copy the result to
        `self.range.array_backend`. Such copies should generally be avoided
        as they can be slow, but they can sometimes be justified if memory
        is scarce on one of the devices.
        """
        return self.__array_backend

    @property
    def device(self):
        """Computational device on which to carry out the BLAS operation.
        See remarks on `array_backend`."""
        return self.__device

    @property
    def axis(self):
        """Axis of domain elements over which is summed."""
        return self.__axis

    @property
    def adjoint(self):
        """Adjoint operator represented by the adjoint matrix.

        Returns
        -------
        adjoint : `MatrixOperator`
        """
        return MatrixOperator(self.matrix.conj().T,
                              domain=self.range, range=self.domain,
                              axis=self.axis)

    @property
    def inverse(self):
        """Inverse operator represented by the inverse matrix.

        Taking the inverse causes sparse matrices to become dense and is
        generally very heavy computationally since the matrix is inverted
        numerically (an O(n^3) operation). It is recommended to instead
        use one of the solvers available in the ``odl.solvers`` package.

        Returns
        -------
        inverse : `MatrixOperator`
        """    
        if self.is_sparse:
            matrix = self._sparse_format.to_dense(self.matrix)
        else:
            matrix = self.matrix
        return MatrixOperator(self.__arr_ns.linalg.inv(matrix),
                                domain=self.range, range=self.domain,
                                axis=self.axis, impl=self.domain.impl, device=self.domain.device)

    def _call(self, x):
        """Return ``self(x[, out])``."""

        if self.is_sparse:
            out = self._sparse_format.matmul_spmatrix_with_vector(self.matrix, x.data)
        else:
            dot = self.__arr_ns.tensordot(self.matrix, x.data, axes=([1], [self.axis]))
            # New axis ends up as first, need to swap it to its place
            out = self.__arr_ns.moveaxis(dot, 0, self.axis)

        return out

    def __repr__(self):
        """Return ``repr(self)``."""
        # Matrix printing itself in an executable way (for dense matrix)
        if self.is_sparse or self.array_backend.impl != 'numpy':
            matrix_str = repr(self.matrix)
        else:
            matrix_str = np.array2string(self.matrix, separator=', ')
        posargs = [matrix_str]

        # Optional arguments with defaults, inferred from the matrix
        range_shape = list(self.domain.shape)
        range_shape[self.axis] = self.matrix.shape[0]

        try:
            default_domain = tensor_space(self.matrix.shape[1],
                                          impl=self.array_backend.impl,
                                          dtype=self.matrix.dtype)
        except (ValueError, TypeError):
            default_domain = None
        try:
            default_range = tensor_space(range_shape,
                                         impl=self.array_backend.impl,
                                         dtype=self.matrix.dtype)
        except (ValueError, TypeError):
            default_range = None

        optargs = [
            ('domain', self.domain, default_domain),
            ('range', self.range, default_range),
            ('axis', self.axis, 0)
        ]

        inner_str = signature_string(posargs, optargs, sep=[', ', ', ', ',\n'],
                                     mod=[['!s'], ['!r', '!r', '']])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(inner_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


def _normalize_sampling_points(sampling_points, ndim):
    """Normalize points to an ndim-long list of linear index arrays.

    This helper converts sampling indices for `SamplingOperator` from
    integers or array-like objects to a list of length ``ndim``, where
    each entry is a `numpy.ndarray` with ``dtype=int``.
    The function also checks if all arrays have equal lengths, and that
    they fulfill ``array.ndim=1`` (or ``size=0`` for if ``ndim == 0``).

    The result of this normalization is intended to be used for indexing
    an ``ndim``-dimensional array at ``sampling_points`` via NumPy fancy
    indexing, i.e., ``result = ndim_array[sampling_points]``.
    """
    sampling_points_in = sampling_points
    if ndim == 0:
        sampling_points = [np.array(sampling_points, dtype=int, copy=AVOID_UNNECESSARY_COPY)]
        if sampling_points[0].size != 0:
            raise ValueError('`sampling_points` must be empty for '
                             '0-dim. `domain`')
    elif ndim == 1:
        if isinstance(sampling_points, Integral):
            sampling_points = (sampling_points,)
        sampling_points = np.array(sampling_points, dtype=int, copy=AVOID_UNNECESSARY_COPY,
                                   ndmin=1)

        # Handle possible list of length one
        if sampling_points.ndim == 2 and sampling_points.shape[0] == 1:
            sampling_points = sampling_points[0]

        sampling_points = [sampling_points]
        if sampling_points[0].ndim > 1:
            raise ValueError('expected 1D index (array), got {}'
                             ''.format(sampling_points_in))
    else:
        try:
            iter(sampling_points)
        except TypeError:
            raise TypeError('`sampling_points` must be a sequence '
                            'for domain with ndim > 1')
        else:
            if np.ndim(sampling_points) == 1:
                sampling_points = [np.array(p, dtype=int)
                                   for p in sampling_points]
            else:
                sampling_points = [
                    np.array(pts, dtype=int, copy=AVOID_UNNECESSARY_COPY, ndmin=1)
                    for pts in sampling_points]
                if any(pts.ndim != 1 for pts in sampling_points):
                    raise ValueError(
                        'index arrays in `sampling_points` must be 1D, '
                        'got {!r}'.format(sampling_points_in))

    return sampling_points


class SamplingOperator(Operator):

    """Operator that samples coefficients.

    The operator is defined by ::

        SamplingOperator(f) == c * f[sampling_points]

    with the weight ``c`` being determined by the variant. By choosing
    ``c = 1``, this operator approximates point evaluations or inner
    products with Dirac deltas, see option ``variant='point_eval'``.
    By choosing ``c = cell_volume``, it approximates the integration of
    ``f`` over the indexed cells, see option ``variant='integrate'``.
    """

    def __init__(self, domain, sampling_points, variant='point_eval'):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `TensorSpace`
            Set of elements on which this operator acts.
        sampling_points : 1D `array-like` or sequence of 1D array-likes
            Indices that determine the sampling points.
            In n dimensions, it should be a sequence of n arrays, where
            each member array is of equal length N. The indexed positions
            are ``(arr1[i], arr2[i], ..., arrn[i])``, in total N
            points.
            If ``domain`` is one-dimensional, a single array-like can be
            used. Likewise, a single point can be given as integer in 1D,
            and as a array-like sequence in nD.
        variant : {'point_eval', 'integrate'}, optional
            For ``'point_eval'`` this operator performs the sampling by
            evaluation the function at the sampling points. The
            ``'integrate'`` variant approximates integration by
            multiplying point evaluation with the cell volume.

        Examples
        --------
        Sampling in 1d can be done with a single index (an int) or a
        sequence of such:

        >>> space = odl.uniform_discr(0, 1, 4)
        >>> op = odl.SamplingOperator(space, sampling_points=1)
        >>> x = space.element([1, 2, 3, 4])
        >>> op(x)
        rn(1).element([ 2.])
        >>> op = odl.SamplingOperator(space, sampling_points=[1, 2, 1])
        >>> op(x)
        rn(3).element([ 2.,  3.,  2.])

        There are two variants ``'point_eval'`` (default) and
        ``'integrate'``, where the latter scales values by the cell
        volume to approximate the integral over the cells of the points:

        >>> op = odl.SamplingOperator(space, sampling_points=[1, 2, 1],
        ...                           variant='integrate')
        >>> space.cell_volume  # the scaling constant
        0.25
        >>> op(x)
        rn(3).element([ 0.5 ,  0.75,  0.5 ])

        In higher dimensions, a sequence of index array-likes must be
        given, or a single sequence for a single point:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (2, 3))
        >>> # Sample at the index (0, 2)
        >>> op = odl.SamplingOperator(space, sampling_points=[0, 2])
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> op(x)
        rn(1).element([ 3.])
        >>> sampling_points = [[0, 1, 1],  # indices (0, 2), (1, 1), (1, 0)
        ...                    [2, 1, 0]]
        >>> op = odl.SamplingOperator(space, sampling_points)
        >>> op(x)
        rn(3).element([ 3.,  5.,  4.])
        """
        if not isinstance(domain, TensorSpace):
            raise TypeError('`domain` must be a `TensorSpace` instance, got '
                            '{!r}'.format(domain))

        self.__sampling_points = _normalize_sampling_points(sampling_points,
                                                            domain.ndim)
        # Flatten indices during init for faster indexing later
        indices_flat = np.ravel_multi_index(self.sampling_points,
                                            dims=domain.shape)
        if np.isscalar(indices_flat):
            self._indices_flat = np.array([indices_flat], dtype=int)
        else:
            self._indices_flat = indices_flat
        self.__variant = str(variant).lower()
        if self.variant not in ('point_eval', 'integrate'):
            raise ValueError('`variant` {!r} not understood'.format(variant))

        # Propagating the impl and device of the range
        ran = tensor_space(self.sampling_points[0].size, dtype=domain.dtype, impl=domain.impl, device=domain.device)
        super(SamplingOperator, self).__init__(domain, ran, linear=True)

    @property
    def variant(self):
        """Weighting scheme for the sampling operator."""
        return self.__variant

    @property
    def sampling_points(self):
        """Indices where to sample the function."""
        return self.__sampling_points

    def _call(self, x):
        """Return values at indices, possibly weighted."""
        out = x.asarray().ravel()[self._indices_flat]

        if self.variant == 'point_eval':
            weights = 1.0
        elif self.variant == 'integrate':
            weights = getattr(self.domain, 'cell_volume', 1.0)
        else:
            raise RuntimeError('bad variant {!r}'.format(self.variant))

        if weights != 1.0:
            out *= weights

        return out

    @property
    def adjoint(self):
        """Adjoint of the sampling operator, a `WeightedSumSamplingOperator`.

        If each sampling point occurs only once, the adjoint consists
        in inserting the given values into the output at the sampling
        points. Duplicate sampling points are weighted with their
        multiplicity.

        Examples
        --------
        >>> space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 3))
        >>> sampling_points = [[0, 1, 1, 0],
        ...                    [0, 1, 2, 0]]
        >>> op = odl.SamplingOperator(space, sampling_points)
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> np.abs(op.adjoint(op(x)).inner(x) - op(x).inner(op(x))) < 1e-10
        True

        The ``'integrate'`` variant adjoint puts ones at the indices in
        ``sampling_points``, multiplied by their multiplicity:

        >>> op = odl.SamplingOperator(space, sampling_points,
        ...                           variant='integrate')
        >>> op.adjoint(op.range.one())  # (0, 0) occurs twice
        uniform_discr([-1., -1.], [ 1.,  1.], (2, 3)).element(
            [[ 2.,  0.,  0.],
             [ 0.,  1.,  1.]]
        )
        >>> np.abs(op.adjoint(op(x)).inner(x) - op(x).inner(op(x))) < 1e-10
        True
        """
        if self.variant == 'point_eval':
            variant = 'dirac'
        elif self.variant == 'integrate':
            variant = 'char_fun'
        else:
            raise RuntimeError('bad variant {!r}'.format(self.variant))

        return WeightedSumSamplingOperator(self.domain, self.sampling_points,
                                           variant)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain, self.sampling_points]
        optargs = [('variant', self.variant, 'point_eval')]
        sig_str = signature_string(posargs, optargs, mod=['!r', ''],
                                   sep=[',\n', '', ',\n'])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(sig_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


class WeightedSumSamplingOperator(Operator):

    r"""Operator computing the sum of coefficients at sampling locations.

    This operator is the adjoint of `SamplingOperator`.

    Notes
    -----
    The weighted sum sampling operator for a sequence
    :math:`I = (i_n)_{n=1}^N`
    of indices (possibly with duplicates) is given by

    .. math::
        W_I(g)(x) = \sum_{i \in I} d_i(x) g_i,

    where :math:`g \in \mathbb{F}^N` is the value vector, and
    :math:`d_i` is either a Dirac delta or a characteristic function of
    the cell centered around the point indexed by :math:`i`.
    """

    def __init__(self, range, sampling_points, variant='char_fun'):
        """Initialize a new instance.

        Parameters
        ----------
        range : `TensorSpace`
            Set of elements into which this operator maps.
        sampling_points : 1D `array-like` or sequence of 1D array-likes
            Indices that determine the sampling points.
            In n dimensions, it should be a sequence of n arrays, where
            each member array is of equal length N. The indexed positions
            are ``(arr1[i], arr2[i], ..., arrn[i])``, in total N
            points.
            If ``range`` is one-dimensional, a single array-like can be
            used. Likewise, a single point can be given as integer in 1D,
            and as a array-like sequence in nD.
        variant : {'char_fun', 'dirac'}, optional
            This option determines which function to sum over.

        Examples
        --------
        In 1d, a single index (an int) or a sequence of such can be used
        for indexing.

        >>> space = odl.uniform_discr(0, 1, 4)
        >>> op = odl.WeightedSumSamplingOperator(space, sampling_points=1)
        >>> op.domain
        rn(1)
        >>> x = op.domain.element([1])
        >>> # Put value 1 at index 1
        >>> op(x)
        uniform_discr(0.0, 1.0, 4).element([ 0.,  1.,  0.,  0.])
        >>> op = odl.WeightedSumSamplingOperator(space,
        ...                                      sampling_points=[1, 2, 1])
        >>> op.domain
        rn(3)
        >>> x = op.domain.element([1, 0.5, 0.25])
        >>> # Index 1 occurs twice and gets two contributions (1 and 0.25)
        >>> op(x)
        uniform_discr(0.0, 1.0, 4).element([ 0.  ,  1.25,  0.5 ,  0.  ])

        The ``'dirac'`` variant scales the values by the reciprocal
        cell volume of the operator range:

        >>> op = odl.WeightedSumSamplingOperator(
        ...     space, sampling_points=[1, 2, 1], variant='dirac')
        >>> x = op.domain.element([1, 0.5, 0.25])
        >>> 1 / op.range.cell_volume  # the scaling constant
        4.0
        >>> op(x)
        uniform_discr(0.0, 1.0, 4).element([ 0.,  5.,  2.,  0.])

        In higher dimensions, a sequence of index array-likes must be
        given, or a single sequence for a single point:

        >>> space = odl.uniform_discr([0, 0], [1, 1], (2, 3))
        >>> # Sample at the index (0, 2)
        >>> op = odl.WeightedSumSamplingOperator(space,
        ...                                      sampling_points=[0, 2])
        >>> x = op.domain.element([1])
        >>> # Insert the value 1 at index (0, 2)
        >>> op(x)
        uniform_discr([ 0.,  0.], [ 1.,  1.], (2, 3)).element(
            [[ 0.,  0.,  1.],
             [ 0.,  0.,  0.]]
        )
        >>> sampling_points = [[0, 1],  # indices (0, 2) and (1, 1)
        ...                    [2, 1]]
        >>> op = odl.WeightedSumSamplingOperator(space, sampling_points)
        >>> x = op.domain.element([1, 2])
        >>> op(x)
        uniform_discr([ 0.,  0.], [ 1.,  1.], (2, 3)).element(
            [[ 0.,  0.,  1.],
             [ 0.,  2.,  0.]]
        )
        """
        if not isinstance(range, TensorSpace):
            raise TypeError('`range` must be a `TensorSpace` instance, got '
                            '{!r}'.format(range))
        self.__sampling_points = _normalize_sampling_points(sampling_points,
                                                            range.ndim)
        # Convert a list of index arrays to linear index array
        indices_flat = np.ravel_multi_index(self.sampling_points,
                                            dims=range.shape)
        if np.isscalar(indices_flat):
            indices_flat = np.array([indices_flat], dtype=int)
        else:
            indices_flat = np.array(indices_flat, dtype=int)

        ### Always converting the indices to the right data type 
        self._indices_flat = range.array_backend.array_constructor(indices_flat, dtype=int, device=range.device)

        self.__variant = str(variant).lower()
        if self.variant not in ('dirac', 'char_fun'):
            raise ValueError('`variant` {!r} not understood'.format(variant))
        
        # Recording the namespace for bincount
        self.namespace = range.array_backend.array_namespace

        # Propagating the impl and device of the range
        domain = tensor_space(self.sampling_points[0].size, dtype=range.dtype, impl=range.impl, device=range.device)
        super(WeightedSumSamplingOperator, self).__init__(
            domain, range, linear=True)

    @property
    def variant(self):
        """Weighting scheme for the operator."""
        return self.__variant

    @property
    def sampling_points(self):
        """Indices where to sample the function."""
        return self.__sampling_points

    def _call(self, x):
        """Sum all values if indices are given multiple times."""
        y = self.namespace.bincount(self._indices_flat, weights=x.data,
                        minlength=self.range.size)

        out = y.reshape(self.range.shape)

        if self.variant == 'dirac':
            weights = getattr(self.range, 'cell_volume', 1.0)
        elif self.variant == 'char_fun':
            weights = 1.0
        else:
            raise RuntimeError('The variant "{!r}" is not yet supported'
                               ''.format(self.variant))

        if weights != 1.0:
            out /= weights

        return out

    @property
    def adjoint(self):
        """Adjoint of this operator, a `SamplingOperator`.

        The ``'char_fun'`` variant of this operator corresponds to the
        ``'integrate'`` sampling operator, and ``'dirac'`` corresponds to
        ``'point_eval'``.

        Examples
        --------
        >>> space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 3))
        >>> # Point (0, 0) occurs twice
        >>> sampling_points = [[0, 1, 1, 0],
        ...                    [0, 1, 2, 0]]
        >>> op = odl.WeightedSumSamplingOperator(space, sampling_points,
        ...                                      variant='dirac')
        >>> y = op.range.element([[1, 2, 3],
        ...                       [4, 5, 6]])
        >>> op.adjoint(y)
        rn(4).element([ 1.,  5.,  6.,  1.])
        >>> x = op.domain.element([1, 2, 3, 4])
        >>> np.abs(op.adjoint(op(x)).inner(x) - op(x).inner(op(x))) < 1e-10
        True
        >>> op = odl.WeightedSumSamplingOperator(space, sampling_points,
        ...                                      variant='char_fun')
        >>> np.abs(op.adjoint(op(x)).inner(x) - op(x).inner(op(x))) < 1e-10
        True
        """
        if self.variant == 'dirac':
            variant = 'point_eval'
        elif self.variant == 'char_fun':
            variant = 'integrate'
        else:
            raise RuntimeError('The variant "{!r}" is not yet supported'
                               ''.format(self.variant))

        return SamplingOperator(self.range, self.sampling_points, variant)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.range, self.sampling_points]
        optargs = [('variant', self.variant, 'char_fun')]
        sig_str = signature_string(posargs, optargs, mod=['!r', ''],
                                   sep=[',\n', '', ',\n'])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(sig_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


class FlatteningOperator(Operator):

    """Operator that reshapes the object as a column vector.

    The operation performed by this operator is ::

        FlatteningOperator(x) == ravel(x)

    The range of this operator is always a `TensorSpace`, i.e., even if
    the domain is a discrete function space.
    """

    def __init__(self, domain):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `TensorSpace`
            Set of elements on which this operator acts.

        Examples
        --------
        >>> space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 3))
        >>> op = odl.FlatteningOperator(space)
        >>> op.range
        rn(6)
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> op(x)
        rn(6).element([ 1.,  2.,  3.,  4.,  5.,  6.])
        """
        if not isinstance(domain, TensorSpace):
            raise TypeError('`domain` must be a `TensorSpace` instance, got '
                            '{!r}'.format(domain))

        range = tensor_space(domain.size, dtype=domain.dtype)
        super(FlatteningOperator, self).__init__(domain, range, linear=True)

    def _call(self, x):
        """Flatten ``x``."""
        return self.range.element(x.data.reshape([self.range.shape[0]]))

    @property
    def adjoint(self):
        """Adjoint of the flattening, a scaled version of the `inverse`.

        Examples
        --------
        >>> space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 4))
        >>> op = odl.FlatteningOperator(space)
        >>> y = op.range.element([1, 2, 3, 4, 5, 6, 7, 8])
        >>> 1 / space.cell_volume  # the scaling factor
        2.0
        >>> op.adjoint(y)
        uniform_discr([-1., -1.], [ 1.,  1.], (2, 4)).element(
            [[  2.,   4.,   6.,   8.],
             [ 10.,  12.,  14.,  16.]]
        )
        >>> x = space.element([[1, 2, 3, 4],
        ...                    [5, 6, 7, 8]])
        >>> np.abs(op.adjoint(op(x)).inner(x) - op(x).inner(op(x))) < 1e-10
        True
        """
        scaling = getattr(self.domain, 'cell_volume', 1.0)
        return 1 / scaling * self.inverse

    @property
    def inverse(self):
        """Operator that reshapes to original shape.

        Examples
        --------
        >>> space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 4))
        >>> op = odl.FlatteningOperator(space)
        >>> y = op.range.element([1, 2, 3, 4, 5, 6, 7, 8])
        >>> op.inverse(y)
        uniform_discr([-1., -1.], [ 1.,  1.], (2, 4)).element(
            [[ 1.,  2.,  3.,  4.],
             [ 5.,  6.,  7.,  8.]]
        )
        >>> op(op.inverse(y)) == y
        True
        """
        op = self
        scaling = getattr(self.domain, 'cell_volume', 1.0)

        class FlatteningOperatorInverse(Operator):

            """Inverse of `FlatteningOperator`.

            This operator reshapes a flat vector back to original shape::

                FlatteningOperatorInverse(x) == reshape(x, orig_shape)
            """

            def __init__(self):
                """Initialize a new instance."""
                super(FlatteningOperatorInverse, self).__init__(
                    op.range, op.domain, linear=True)

            def _call(self, x):
                """Reshape ``x`` back to n-dim. shape."""
                return np.reshape(x.asarray(), self.range.shape)

            @property
            def adjoint(self):
                """Adjoint of this operator, a scaled `FlatteningOperator`."""
                return scaling * op

            @property
            def inverse(self):
                """Inverse of this operator."""
                return op

            def __repr__(self):
                """Return ``repr(self)``."""
                return '{!r}.inverse'.format(op)

            def __str__(self):
                """Return ``str(self)``."""
                return repr(self)

        return FlatteningOperatorInverse()

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain]
        sig_str = signature_string(posargs, optargs, mod=['!r', ''],
                                   sep=['', '', ',\n'])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(sig_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


def is_compatible_space(space, base_space):
    """Check compatibility of a (power) space with a base space.

    Compatibility here means that the spaces are equal or ``space``
    is a non-empty power space of ``base_space`` up to different
    data types.

    Parameters
    ----------
    space, base_space : `LinearSpace`
        Spaces to check for compatibility. ``base_space`` cannot be a
        `ProductSpace`.

    Returns
    -------
    is_compatible : bool
        ``True`` if

        - ``space == base_space`` or
        - ``space.astype(base_space.dtype) == base_space``, provided that
          these properties exist, or
        - ``space`` is a power space of nonzero length and one of the three
          situations applies to ``space[0]`` (recursively).

        Otherwise ``False``.

    Examples
    --------
    Scalar spaces:

    >>> base = odl.rn(2)
    >>> is_compatible_space(odl.rn(2), base)
    True
    >>> is_compatible_space(odl.rn(3), base)
    False
    >>> is_compatible_space(odl.rn(2, dtype='float32'), base)
    True

    Power spaces:

    >>> is_compatible_space(odl.rn(2) ** 2, base)
    True
    >>> is_compatible_space(odl.rn(2) * odl.rn(3), base)  # no power space
    False
    >>> is_compatible_space(odl.rn(2, dtype='float32') ** 2, base)
    True
    """
    if isinstance(base_space, ProductSpace):
        return False

    if isinstance(space, ProductSpace):
        if not space.is_power_space:
            return False
        elif len(space) == 0:
            return False
        else:
            return is_compatible_space(space[0], base_space)
    else:
        if hasattr(space, 'astype') and hasattr(base_space, 'dtype'):
            # TODO: maybe only the shape should play a role?
            comp_space = space.astype(base_space.dtype)
        else:
            comp_space = space

        return comp_space == base_space


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utility functions for FOMs."""

import numpy as np

from odl.core.discr import uniform_discr
from odl.trafos.backends import PYFFTW_AVAILABLE

__all__ = ()


def filter_image_sep2d(image, fh, fv, impl='numpy', padding=None):
    """Filter an image with a separable filter.

    Parameters
    ----------
    image : 2D array-like
        The image to be filtered. It must have a real (vs. complex) dtype.
    fh, fv : 1D array-like
        Horizontal (axis 0) and vertical (axis 1) filters. Their sizes
        can be at most the image sizes in the respective axes.
    impl : {'numpy', 'pyfftw'}, optional
        FFT backend to use. The ``pyfftw`` backend requires the
        ``pyfftw`` package to be installed. It is usually significantly
        faster than the NumPy backend.
    padding : positive int, optional
        Amount of zeros added to the left and right of the image in all
        axes before FFT. This helps avoiding wraparound artifacts due to
        large boundary values.
        For ``None``, the padding is computed as ::

            padding = min(max(len(fh), len(fv)) - 1, 64)

        A padding of ``len(filt) - 1`` ensures that errors in FFT-based
        convolutions are small. At the same time, the padding should not
        be excessive to retain efficiency.

    Returns
    -------
    filtered : 2D `numpy.ndarray`
        The image filtered horizontally by ``fh`` and vertically by ``fv``.
        It has the same shape as ``image``, and its dtype is
        ``np.result_type(image, fh, fv)``.
    """
    # TODO: generalize for nD
    impl, impl_in = str(impl).lower(), impl
    if impl not in ('numpy', 'pyfftw'):
        raise ValueError('`impl` {!r} not understood'
                         ''.format(impl_in))

    image = np.asarray(image)
    if image.ndim != 2:
        raise ValueError('`image` must be 2-dimensional, got image with '
                         'ndim={}'.format(image.ndim))
    if image.size == 0:
        raise ValueError('`image` cannot have size 0')
    if not np.issubdtype(image.dtype, np.floating):
        image = image.astype(float)

    fh = np.asarray(fh).astype(image.dtype)
    if fh.ndim != 1:
        raise ValueError('`fh` must be one-dimensional')
    elif fh.size == 0:
        raise ValueError('`fh` cannot have size 0')
    elif fh.size > image.shape[0]:
        raise ValueError('`fh` can be at most `image.shape[0]`, got '
                         '{} > {}'.format(fh.size, image.shape[0]))

    fv = np.asarray(fv).astype(image.dtype)
    if fv.ndim != 1:
        raise ValueError('`fv` must be one-dimensional')
    elif fv.size == 0:
        raise ValueError('`fv` cannot have size 0')
    elif fv.size > image.shape[0]:
        raise ValueError('`fv` can be at most `image.shape[1]`, got '
                         '{} > {}'.format(fv.size, image.shape[1]))

    # Pad image with zeros
    if padding is None:
        padding = min(max(len(fh), len(fv)) - 1, 64)

    if padding != 0:
        image_padded = np.pad(image, padding, mode='constant')
    else:
        image_padded = image.copy() if impl == 'pyfftw' else image

    # Prepare filters for the convolution
    def prepare_for_fft(filt, n_new):
        """Return padded and shifted filter ready for FFT.

        The filter is padded with zeros to the new size, and then shifted
        such that such that the middle element of old filter, i.e., the
        one at index ``(len(filt) - 1) // 2`` ends up at index 0.
        """
        mid = (len(filt) - 1) // 2
        padded = np.zeros(n_new, dtype=filt.dtype)
        padded[:len(filt) - mid] = filt[mid:]
        padded[len(padded) - mid:] = filt[:mid]
        return padded

    fh = prepare_for_fft(fh, image_padded.shape[0])
    fv = prepare_for_fft(fv, image_padded.shape[1])

    # Perform the multiplication in Fourier space and apply inverse FFT
    if impl == 'numpy':
        image_ft = np.fft.rfftn(image_padded)
        fh_ft = np.fft.fft(fh)
        fv_ft = np.fft.rfft(fv)

        image_ft *= fh_ft[:, None]
        image_ft *= fv_ft[None, :]
        # Important to specify the shape since `irfftn` cannot know the
        # original shape
        conv = np.fft.irfftn(image_ft, s=image_padded.shape)
        if conv.dtype != image.dtype:
            conv = conv.astype(image.dtype)

    elif impl == 'pyfftw':
        if not PYFFTW_AVAILABLE:
            raise ValueError(
                '`pyfftw` package is not available; you need to install it '
                'to use the pyfftw backend')

        import pyfftw
        import multiprocessing

        # Generate output arrays, for half-complex transform of image and
        # vertical filter, and full FT of the horizontal filter
        out_img_shape = (image_padded.shape[0], image_padded.shape[1] // 2 + 1)
        out_img_dtype = np.result_type(image_padded, 1j)
        out_img = np.empty(out_img_shape, out_img_dtype)

        out_fh_shape = out_img_shape[0]
        out_fh_dtype = np.result_type(fh, 1j)
        fh_c = fh.astype(out_fh_dtype)  # need to make this a C2C trafo
        out_fh = np.empty(out_fh_shape, out_fh_dtype)

        out_fv_shape = out_img_shape[1]
        out_fv_dtype = np.result_type(fv, 1j)
        out_fv = np.empty(out_fv_shape, out_fv_dtype)

        # Perform the forward transforms of image and filters. We use
        # the `FFTW_ESTIMATE` flag to not allow the planner to destroy
        # the input.
        plan = pyfftw.FFTW(image_padded, out_img, axes=(0, 1),
                           direction='FFTW_FORWARD',
                           flags=['FFTW_ESTIMATE'],
                           threads=multiprocessing.cpu_count())
        plan(image_padded, out_img)

        plan = pyfftw.FFTW(fh_c, out_fh, axes=(0,),
                           direction='FFTW_FORWARD',
                           flags=['FFTW_ESTIMATE'],
                           threads=multiprocessing.cpu_count())
        plan(fh_c, out_fh)

        plan = pyfftw.FFTW(fv, out_fv, axes=(0,),
                           direction='FFTW_FORWARD',
                           flags=['FFTW_ESTIMATE'],
                           threads=multiprocessing.cpu_count())
        plan(fv, out_fv)

        # Fourier space multiplication
        out_img *= out_fh[:, None]
        out_img *= out_fv[None, :]

        # Inverse trafo
        conv = image_padded  # Overwrite
        plan = pyfftw.FFTW(out_img.copy(), conv, axes=(0, 1),
                           direction='FFTW_BACKWARD',
                           flags=['FFTW_ESTIMATE'],
                           threads=multiprocessing.cpu_count())
        plan(out_img, conv)

    else:
        raise ValueError('unsupported `impl` {!r}'.format(impl_in))

    if padding:
        return conv[padding:-padding, padding:-padding]
    else:
        return conv


def haarpsi_similarity_map(img1, img2, axis, c, a):
    r"""Local similarity map for directional features along an axis.

    Parameters
    ----------
    img1, img2 : array-like
        The images to compare. They must have equal shape.
    axis : {0, 1}
        Direction in which to look for edge similarities.
    c : positive float
        Constant determining the score of maximally dissimilar values.
        Smaller constant means higher penalty for dissimilarity.
        See Notes for details.
    a : positive float
        Parameter in the logistic function. Larger value leads to a
        steeper curve, thus lowering the threshold for an input to
        be mapped to an output close to 1. See Notes for details.

    Returns
    -------
    local_sim : `numpy.ndarray`
        Pointwise similarity of directional edge features of ``img1`` and
        ``img2``, measured using two Haar wavelet detail levels.

    Notes
    -----
    For input images :math:`f_1, f_2` this function is defined as

    .. math::
        \mathrm{HS}_{f_1, f_2}^{(k)}(x) =
        l_a \left(
        \frac{1}{2} \sum_{j=1}^2
        S\left(\left|g_j^{(k)} \ast f_1 \right|(x),
        \left|g_j^{(k)} \ast f_2 \right|(x), c\right)
        \right),

    see `[Rei+2016] <https://arxiv.org/abs/1607.06140>`_ equation (10).
    Here, the superscript :math:`(k)` refers to the axis (0 or 1)
    in which edge features are compared, :math:`l_a` is the logistic
    function :math:`l_a(x) = (1 + \mathrm{e}^{-a x})^{-1}`, and :math:`S`
    is the pointwise similarity score

    .. math::
        S(x, y, c) = \frac{2xy + c^2}{x^2 + y^2 + c^2},

    Hence, :math:`c` is the :math:`y`-value at which the score
    drops to :math:`1 / 2` for :math:`x = 0`. In other words, the smaller
    :math:`c` is chosen, the more dissimilarity is penalized.

    The filters :math:`g_j^{(k)}` are high-pass Haar wavelet filters in the
    axis :math:`k` and low-pass Haar wavelet filters in the other axes.
    The index :math:`j` refers to the scaling level of the wavelet.
    In code, these filters can be computed as ::

        f_lo_level1 = [np.sqrt(2), np.sqrt(2)] # low-pass Haar filter
        f_hi_level1 = [-np.sqrt(2), np.sqrt(2)] # high-pass Haar filter
        f_lo_level2 = np.repeat(f_lo_level1, 2)
        f_hi_level2 = np.repeat(f_hi_level1, 2)
        f_lo_level3 = np.repeat(f_lo_level2, 2)
        f_hi_level3 = np.repeat(f_hi_level2, 2)
        ...

    The logistic function :math:`l_a` transforms values in
    :math:`[0, \infty)` to :math:`[1/2, 1)`, where the parameter
    :math:`a` determines how fast the curve attains values close
    to 1. Larger :math:`a` means that smaller :math:`x` will yield
    a value :math:`l_a(x)` close to 1 (and thus result in a higher
    score). In other words, the larger :math:`a`, the more forgiving
    the similarity measure.

    References
    ----------
    [Rei+2016] Reisenhofer, R, Bosse, S, Kutyniok, G, and Wiegand, T.
    *A Haar Wavelet-Based Perceptual Similarity Index for Image Quality
    Assessment*. arXiv:1607.06140 [cs], Jul. 2016.
    """
    # TODO: generalize for nD
    import scipy.special
    impl = 'pyfftw' if PYFFTW_AVAILABLE else 'numpy'

    # Haar wavelet filters for levels 1 and 2
    dec_lo_lvl1 = np.array([np.sqrt(2), np.sqrt(2)])
    dec_lo_lvl2 = np.repeat(dec_lo_lvl1, 2)
    dec_hi_lvl1 = np.array([-np.sqrt(2), np.sqrt(2)])
    dec_hi_lvl2 = np.repeat(dec_hi_lvl1, 2)

    if axis == 0:
        # High-pass in axis 0, low-pass in axis 1
        fh_lvl1 = dec_hi_lvl1
        fv_lvl1 = dec_lo_lvl1
        fh_lvl2 = dec_hi_lvl2
        fv_lvl2 = dec_lo_lvl2
    elif axis == 1:
        # Low-pass in axis 0, high-pass in axis 1
        fh_lvl1 = dec_lo_lvl1
        fv_lvl1 = dec_hi_lvl1
        fh_lvl2 = dec_lo_lvl2
        fv_lvl2 = dec_hi_lvl2
    else:
        raise ValueError('`axis` out of the valid range 0 -> 1')

    # Filter images with level 1 and 2 filters
    img1_lvl1 = filter_image_sep2d(img1, fh_lvl1, fv_lvl1, impl=impl)
    img1_lvl2 = filter_image_sep2d(img1, fh_lvl2, fv_lvl2, impl=impl)

    img2_lvl1 = filter_image_sep2d(img2, fh_lvl1, fv_lvl1, impl=impl)
    img2_lvl2 = filter_image_sep2d(img2, fh_lvl2, fv_lvl2, impl=impl)

    c = float(c)

    def S(x, y):
        """Return ``(2 * x * y + c ** 2) / (x ** 2 + y ** 2 + c ** 2)``."""
        num = 2 * x
        num *= y
        num += c ** 2
        denom = x ** 2
        denom += y ** 2
        denom += c ** 2
        frac = num
        frac /= denom
        return frac

    # Compute similarity scores for both levels
    np.abs(img1_lvl1, out=img1_lvl1)
    np.abs(img2_lvl1, out=img2_lvl1)
    np.abs(img1_lvl2, out=img1_lvl2)
    np.abs(img2_lvl2, out=img2_lvl2)

    sim_lvl1 = S(img1_lvl1, img2_lvl1)
    sim_lvl2 = S(img1_lvl2, img2_lvl2)

    # Return logistic of the mean value
    sim = sim_lvl1
    sim += sim_lvl2
    sim /= 2
    sim *= a
    return scipy.special.expit(sim)


def haarpsi_weight_map(img1, img2, axis):
    r"""Weighting map for directional features along an axis.

    Parameters
    ----------
    img1, img2 : array-like
        The images to compare. They must have equal shape.
    axis : {0, 1}
        Direction in which to look for edge similarities.

    Returns
    -------
    weight_map : `numpy.ndarray`
        The pointwise weight map. See Notes for details.

    Notes
    -----
    The pointwise weight map of associated with input images :math:`f_1, f_2`
    and axis :math:`k` is defined
    as

    .. math::
        \mathrm{W}_{f_1, f_2}^{(k)}(x) =
        \max \left\{
            \left|g_3^{(k)} \ast f_1 \right|(x),
            \left|g_3^{(k)} \ast f_2 \right|(x)
        \right\},

    see `[Rei+2016] <https://arxiv.org/abs/1607.06140>`_ equations (11)
    and (13).

    Here, :math:`g_3^{(k)}` is a Haar wavelet filter for scaling level 3
    that performs high-pass filtering in axis :math:`k` and low-pass
    filtering in the other axes. Such a filter can be computed as ::

        f_lo_level1 = [np.sqrt(2), np.sqrt(2)] # low-pass Haar filter
        f_hi_level1 = [-np.sqrt(2), np.sqrt(2)] # high-pass Haar filter
        f_lo_level3 = np.repeat(f_lo_level1, 4)
        f_hi_level3 = np.repeat(f_hi_level1, 4)

    References
    ----------
    [Rei+2016] Reisenhofer, R, Bosse, S, Kutyniok, G, and Wiegand, T.
    *A Haar Wavelet-Based Perceptual Similarity Index for Image Quality
    Assessment*. arXiv:1607.06140 [cs], Jul. 2016.
    """
    # TODO: generalize for nD
    impl = 'pyfftw' if PYFFTW_AVAILABLE else 'numpy'

    # Haar wavelet filters for level 3
    dec_lo_lvl3 = np.repeat([np.sqrt(2), np.sqrt(2)], 4)
    dec_hi_lvl3 = np.repeat([-np.sqrt(2), np.sqrt(2)], 4)

    if axis == 0:
        fh_lvl3 = dec_hi_lvl3
        fv_lvl3 = dec_lo_lvl3
    elif axis == 1:
        fh_lvl3 = dec_lo_lvl3
        fv_lvl3 = dec_hi_lvl3
    else:
        raise ValueError('`axis` out of the valid range 0 -> 1')

    # Filter with level 3 wavelet filter
    img1_lvl3 = filter_image_sep2d(img1, fh_lvl3, fv_lvl3, impl=impl)
    img2_lvl3 = filter_image_sep2d(img2, fh_lvl3, fv_lvl3, impl=impl)

    # Return the pointwise maximum of the filtered images
    np.abs(img1_lvl3, out=img1_lvl3)
    np.abs(img2_lvl3, out=img2_lvl3)

    return np.maximum(img1_lvl3, img2_lvl3)


def spherical_sum(image, binning_factor=1.0):
    """Sum image values over concentric annuli.

    Parameters
    ----------
    image : `DiscretizedSpace` element
        Input data whose radial sum should be computed.
    binning_factor : positive float, optional
        Reduce the number of output bins by this factor. Increasing this
        number can help reducing fluctuations due to the variance of points
        that fall in a particular annulus.
        A binning factor of ``1`` corresponds to a bin size equal to
        image pixel size for images with square pixels, otherwise ::

            max(norm2(c)) / norm2(shape)

        where the maximum is taken over all corners of the image domain.

    Returns
    -------
    spherical_sum : 1D `DiscretizedSpace` element
        The spherical sum of ``image``. Its space is one-dimensional with
        domain ``[0, rmax]``, where ``rmax`` is the radius of the smallest
        ball containing ``image.space.domain``. Its shape is ``(N,)`` with ::

            N = int(sqrt(sum(n ** 2 for n in image.shape)) / binning_factor)
    """
    r = np.sqrt(sum(xi ** 2 for xi in image.space.meshgrid))
    rmax = max(np.linalg.norm(c) for c in image.space.domain.corners())
    n_bins = int(np.sqrt(sum(n ** 2 for n in image.shape)) / binning_factor)
    rad_sum, _ = np.histogram(r, weights=image, bins=n_bins, range=(0, rmax))

    out_spc = uniform_discr(min_pt=0, max_pt=rmax, shape=n_bins,
                            impl=image.space.impl, dtype=image.space.dtype,
                            interp="linear", axis_labels=["$r$"])

    return out_spc.element(rad_sum)
